import typing

import pytest
import pytest_twisted
from zope.interface import implementer
from twisted.internet.task import deferLater
from twisted.internet.protocol import Factory
from twisted.internet.interfaces import IProtocolFactory

from ... import create
from ...errors import LonelyError
from ...eventual import EventualQueue
from ..._dilation._noise import NoiseConnection
from ..._status import (
    Disconnected, ConnectedPeer, Closed, ConnectionStatus, PeerSharedKey,
    CodeStatus, PeerConnection
)


def _union_sort_order(union_klass):
    """
    Convert a Union type into a dict mapping its members to a number
    representing their order in the original type
    """
    # note: we want these to be in the same order as in the _status
    # Union-type declarations
    return {
        klass: idx
        for idx, klass in enumerate(typing.get_args(union_klass))
    }


def assert_mailbox_status_order(status_messages):
    """
    Confirm the given status messages have the correct properties.

    Here we want to assert that individual status items "go the right
    way": that is, for most of these things we know that you can't go
    back to "AllegedSharedKey" after seeing a "ConfirmedKey" (for
    example).

    So what we do is make a sort-order for each status field, sort the
    messages by that, and assert that this sorted order is the same as
    the raw message order.
    """
    code_sorting = _union_sort_order(CodeStatus)
    key_sorting = _union_sort_order(PeerSharedKey)
    mailbox_sorting = _union_sort_order(ConnectionStatus)

    code_messages = [st.code for st in status_messages]
    acceptable_order = sorted(code_messages, key=lambda code: code_sorting[type(code)])
    assert acceptable_order == code_messages, "'code' status came in an illegal order"

    key_messages = [st.peer_key for st in status_messages]
    assert sorted(key_messages, key=lambda k: key_sorting[type(k)]) == key_messages

    # "in general" we can go from Connected back to Connecting,
    # but for this particular test we don't actually do that
    mailbox_messages = [st.mailbox_connection for st in status_messages]
    assert sorted(mailbox_messages, key=lambda k: mailbox_sorting[type(k)]) == mailbox_messages
    # initial and terminal status must be correct
    assert isinstance(mailbox_messages[0], Disconnected)
    assert isinstance(mailbox_messages[-1], Closed)


def assert_dilation_status_order(status_messages):
    """
    for "Dilation"-specific messages, we only analyze the non-wormhole
    status parts and use a similar trick to the above.
    """
    generations = [msg.generation for msg in status_messages]
    # generations must always increase ("sorted" is a stable-sort .. right??)
    assert sorted(generations) == generations, "generation number went backwards"

    # "In general" the peer_connection can go ConnectedPeer back to
    # ConnectingPeer multiple times, but we don't actually do that in
    # this particular test.
    #
    # shapr points out we could fix this by setting ConnectedPeer and
    # ReconnectingPeer to the same number -- then they should
    # stable-sort to the right thing (however, we do not yet have a
    # test that actually re-connects a peer). e.g.:
    #     peer_sorting[ReconnectingPeer] = peer_sorting[ConnectedPeer]
    peer_sorting = _union_sort_order(PeerConnection)
    peers = [msg.peer_connection for msg in status_messages]
    assert sorted(peers, key=lambda k: peer_sorting[type(k)]) == peers, "peer status went backwards"


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_on_status_error(reactor, mailbox):
    """
    Our user code raises an exception during status processing
    """
    eq = EventualQueue(reactor)

    class FakeError(Exception):
        pass

    def on_status(_):
        raise FakeError()
    with pytest.raises(FakeError):
        w = create(
            "appid", mailbox.url,
            reactor,
            versions={"fun": "quux"},
            _eventual_queue=eq,
            dilation=True,
            on_status_update=on_status,
        )
        await w.allocate_code()
        code = await w.get_code()
        print(code)
        try:
            await w.close()
        except LonelyError:
            pass


@implementer(IProtocolFactory)
class SubFac(Factory):
    subprotocol = "jemison"


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_dilation_status(reactor, mailbox):
    eq = EventualQueue(reactor)

    status0 = []
    status1 = []

    wormhole_status0 = []
    wormhole_status1 = []

    w0 = create(
        "appid", mailbox.url,
        reactor,
        versions={"fun": "quux"},
        _eventual_queue=eq,
        dilation=True,
        on_status_update=wormhole_status0.append,
    )

    w1 = create(
        "appid", mailbox.url,
        reactor,
        versions={"bar": "baz"},
        _eventual_queue=eq,
        dilation=True,
        on_status_update=wormhole_status1.append,
    )

    w0.allocate_code()
    code = await w0.get_code()

    w1.set_code(code)

    w0.dilate(on_status_update=status0.append)
    w1.dilate(on_status_update=status1.append)

    # we should see the _other side's_ app-versions
    v0 = await w1.get_versions()
    v1 = await w0.get_versions()
    assert v0 == {"fun": "quux"}
    assert v1 == {"bar": "baz"}

    @pytest_twisted.ensureDeferred()
    async def wait_for_peer():
        while True:
            await deferLater(reactor, 0.001, lambda: None)
            peers = [
                st
                for st in status0
                if isinstance(st.peer_connection, ConnectedPeer)
            ]
            if peers:
                return
    await wait_for_peer()

    # we don't actually do anything, just disconnect after we have
    # our peer
    await w0.close()
    await w1.close()

    # analyze the message orders

    assert_mailbox_status_order(wormhole_status0)
    assert_mailbox_status_order(wormhole_status1)

    assert_dilation_status_order(status0)
    assert_dilation_status_order(status1)

    # todo: we could expand these to timestamps: we know they
    # should never go backwards



# This is a relay I run on a personal server. If it gets too expensive to
# run, I'll shut it down.
MAILBOX_RELAY = RENDEZVOUS_RELAY = "ws://relay.magic-wormhole.io:4000/v1"
TRANSIT_RELAY = "tcp:transit.magic-wormhole.io:4001"



import traceback
from sys import stderr

from attr import attrib, attrs
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import blockingCallFromThread, deferToThread

from .errors import AlreadyInputNameplateError, KeyFormatError

try:
    import readline
except ImportError:
    readline = None

errf = None


# uncomment this to enable tab-completion debugging
# import os ; errf = open("err", "w") if os.path.exists("err") else None
def debug(*args, **kwargs):  # pragma: no cover
    if errf:
        print(*args, file=errf, **kwargs)
        errf.flush()


@attrs
class CodeInputter:
    _input_helper = attrib()
    _reactor = attrib()

    def __attrs_post_init__(self):
        self.used_completion = False
        self._matches = None
        # once we've claimed the nameplate, we can't go back
        self._committed_nameplate = None  # or string

    def bcft(self, f, *a, **kw):
        return blockingCallFromThread(self._reactor, f, *a, **kw)

    def completer(self, text, state):
        try:
            return self._wrapped_completer(text, state)
        except Exception as e:
            # completer exceptions are normally silently discarded, which
            # makes debugging challenging
            print(f"completer exception: {e}")
            traceback.print_exc()
            raise

    def _wrapped_completer(self, text, state):
        self.used_completion = True
        # if we get here, then readline must be active
        ct = readline.get_completion_type()
        if state == 0:
            debug("completer starting (%s) (state=0) (ct=%d)" % (text, ct))
            self._matches = self._commit_and_build_completions(text)
            debug(" matches:", " ".join([f"'{m}'" for m in self._matches]))
        else:
            debug(" s%d t'%s' ct=%d" % (state, text, ct))

        if state >= len(self._matches):
            debug("  returning None")
            return None
        debug(f"  returning '{self._matches[state]}'")
        return self._matches[state]

    def _commit_and_build_completions(self, text):
        ih = self._input_helper
        if "-" in text:
            got_nameplate = True
            nameplate, words = text.split("-", 1)
        else:
            got_nameplate = False
            nameplate = text  # partial

        # 'text' is one of these categories:
        #  "" or "12": complete on nameplates (all that match, maybe just one)

        #  "123-": if we haven't already committed to a nameplate, commit and
        #  wait for the wordlist. Then (either way) return the whole wordlist.

        #  "123-supp": if we haven't already committed to a nameplate, commit
        #  and wait for the wordlist. Then (either way) return all current
        #  matches.

        if self._committed_nameplate:
            if not got_nameplate or nameplate != self._committed_nameplate:
                # they deleted past the commitment point: we can't use
                # this. For now, bail, but in the future let's find a
                # gentler way to encourage them to not do that.
                raise AlreadyInputNameplateError(
                    "nameplate (%s-) already entered, cannot go back" %
                    self._committed_nameplate)
        if not got_nameplate:
            # we're completing on nameplates: "" or "12" or "123"
            self.bcft(ih.refresh_nameplates)  # results arrive later
            debug("  getting nameplates")
            completions = self.bcft(ih.get_nameplate_completions, nameplate)
        else:  # "123-" or "123-supp"
            # time to commit to this nameplate, if they haven't already
            if not self._committed_nameplate:
                debug(f"  choose_nameplate({nameplate})")
                self.bcft(ih.choose_nameplate, nameplate)
                self._committed_nameplate = nameplate

                # Now we want to wait for the wordlist to be available. If
                # the user just typed "12-supp TAB", we'll claim "12" but
                # will need a server roundtrip to discover that "supportive"
                # is the only match. If we don't block, we'd return an empty
                # wordlist to readline (which will beep and show no
                # completions). *Then* when the user hits TAB again a moment
                # later (after the wordlist has arrived, but the user hasn't
                # modified the input line since the previous empty response),
                # readline would show one match but not complete anything.

                # In general we want to avoid returning empty lists to
                # readline. If the user hits TAB when typing in the nameplate
                # (before the sender has established one, or before we're
                # heard about it from the server), it can't be helped. But
                # for the rest of the code, a simple wait-for-wordlist will
                # improve the user experience.
                self.bcft(ih.when_wordlist_is_available)  # blocks on CLAIM
            # and we're completing on words now
            debug(f"  getting words ({words})")
            completions = [
                nameplate + "-" + c
                for c in self.bcft(ih.get_word_completions, words)
            ]

        # rlcompleter wants full strings
        return sorted(completions)

    def finish(self, text):
        if "-" not in text:
            raise KeyFormatError("incomplete wormhole code")
        nameplate, words = text.split("-", 1)

        if self._committed_nameplate:
            if nameplate != self._committed_nameplate:
                # they deleted past the commitment point: we can't use
                # this. For now, bail, but in the future let's find a
                # gentler way to encourage them to not do that.
                raise AlreadyInputNameplateError(
                    "nameplate (%s-) already entered, cannot go back" %
                    self._committed_nameplate)
        else:
            debug(f"  choose_nameplate({nameplate})")
            self.bcft(self._input_helper.choose_nameplate, nameplate)
        debug(f"  choose_words({words})")
        self.bcft(self._input_helper.choose_words, words)


def _input_code_with_completion(prompt, input_helper, reactor):
    # reminder: this all occurs in a separate thread. All calls to input_helper
    # must go through blockingCallFromThread()
    c = CodeInputter(input_helper, reactor)
    if readline is not None:
        if readline.__doc__ and "libedit" in readline.__doc__:
            readline.parse_and_bind("bind ^I rl_complete")
        else:
            readline.parse_and_bind("tab: complete")
        readline.set_completer(c.completer)
        readline.set_completer_delims("")
        debug("==== readline-based completion is prepared")
    else:
        debug("==== unable to import readline, disabling completion")
    code = input(prompt)
    # Code is str(bytes) on py2, and str(unicode) on py3. We want unicode.
    if isinstance(code, bytes):
        code = code.decode("utf-8")
    c.finish(code)
    return c.used_completion


def warn_readline():
    # When our process receives a SIGINT, Twisted's SIGINT handler will
    # stop the reactor and wait for all threads to terminate before the
    # process exits. However, if we were waiting for
    # input_code_with_completion() when SIGINT happened, the readline
    # thread will be blocked waiting for something on stdin. Trick the
    # user into satisfying the blocking read so we can exit.
    print("\nCommand interrupted: please press Return to quit", file=stderr)

    # Other potential approaches to this problem:
    # * hard-terminate our process with os._exit(1), but make sure the
    #   tty gets reset to a normal mode ("cooked"?) first, so that the
    #   next shell command the user types is echoed correctly
    # * track down the thread (t.p.threadable.getThreadID from inside the
    #   thread), get a cffi binding to pthread_kill, deliver SIGINT to it
    # * allocate a pty pair (pty.openpty), replace sys.stdin with the
    #   slave, build a pty bridge that copies bytes (and other PTY
    #   things) from the real stdin to the master, then close the slave
    #   at shutdown, so readline sees EOF
    # * write tab-completion and basic editing (TTY raw mode,
    #   backspace-is-erase) without readline, probably with curses or
    #   twisted.conch.insults
    # * write a separate program to get codes (maybe just "wormhole
    #   --internal-get-code"), run it as a subprocess, let it inherit
    #   stdin/stdout, send it SIGINT when we receive SIGINT ourselves. It
    #   needs an RPC mechanism (over some extra file descriptors) to ask
    #   us to fetch the current nameplate_id list.
    #
    # Note that hard-terminating our process with os.kill(os.getpid(),
    # signal.SIGKILL), or SIGTERM, doesn't seem to work: the thread
    # doesn't see the signal, and we must still wait for stdin to make
    # readline finish.


@inlineCallbacks
def input_with_completion(prompt, input_helper, reactor):
    t = reactor.addSystemEventTrigger("before", "shutdown", warn_readline)
    # input_helper.refresh_nameplates()
    used_completion = yield deferToThread(_input_code_with_completion, prompt,
                                          input_helper, reactor)
    reactor.removeSystemEventTrigger(t)
    return used_completion



import wormhole
from zope.interface import implementer, alsoProvides
from twisted.internet.defer import Deferred, gatherResults
from twisted.internet.protocol import Protocol, Factory
from twisted.internet.interfaces import IProtocolFactory

import pytest
import pytest_twisted

from ..common import poll_until
from ..._interfaces import IDilationConnector
from ...eventual import EventualQueue
from ..._dilation._noise import NoiseConnection


APPID = "lothar.com/dilate-test"


def doBoth(d1, d2):
    return gatherResults([d1, d2], True)


class HelloProtocol(Protocol):
    def connectionMade(self):
        print("got connection")
        self.transport.write(b"hello\n")

    def dataReceived(self, data):
        print(f"dataReceived: {data}")
        self.factory.d.callback(data)

    def connectionLost(self, why):
        print("connectionLost")


class BonjourProtocol(Protocol):
    """
    A french version of HelloProtocol
    """
    def connectionMade(self):
        self.transport.write(b"bonjour\n")

    def dataReceived(self, data):
        self.factory.d.callback(data)

    def connectionLost(self, why):
        pass


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_single_subprotocol(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w2 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w1.allocate_code()
    code = await w1.get_code()
    print(f"code is: {code}")
    w2.set_code(code)
    await doBoth(w1.get_verifier(), w2.get_verifier())
    print("connected")

    # side "0" is the host / listener, side "1" is the guest / connector
    fserv0 = Factory()
    fserv0.d = Deferred()
    fserv0.protocol = HelloProtocol

    eps1 = w1.dilate()
    eps2 = w2.dilate()
    print("w.dilate ready")

    eps1.listener_for("proto").listen(fserv0)

    f2 = Factory()
    alsoProvides(f2, IProtocolFactory)
    f2.protocol = HelloProtocol
    f2.d = Deferred()
    f2.d.addCallback(lambda data: eq.fire_eventually(data))
    d2 = eps2.connector_for("proto").connect(f2)
    await d2

    # note: I'm making some horrible assumptions about one-to-one writes
    # and reads across a TCP stack that isn't obligated to maintain such
    # a relationship, but it's much easier than doing this properly. If
    # the tests ever start failing, do the extra work, probably by
    # using a twisted.protocols.basic.LineOnlyReceiver
    data1 = await fserv0.d
    data2 = await f2.d
    assert data1 == b"hello\n"
    assert data2 == b"hello\n"

    await w2.close()


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_double_subprotocol(reactor, mailbox):
    """
    Two subprotocols with different names / implementations on the
    same wormhole.
    """
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w2 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w1.allocate_code()
    code = await w1.get_code()
    print(f"code is: {code}")
    w2.set_code(code)
    await doBoth(w1.get_verifier(), w2.get_verifier())
    print("connected")

    # side "0" is the host / listener, side "1" is the guest / connector
    fserv0 = Factory()
    fserv0.d = Deferred()
    fserv0.protocol = HelloProtocol

    fserv1 = Factory()
    fserv1.d = Deferred()
    fserv1.protocol = BonjourProtocol

    host = w1.dilate()
    guest = w2.dilate()

    host.listener_for("hello").listen(fserv0)
    host.listener_for("bonjour").listen(fserv1)

    f2 = Factory.forProtocol(HelloProtocol)
    f2.d = Deferred()
    f2.d.addCallback(lambda data: eq.fire_eventually(data))

    f3 = Factory.forProtocol(BonjourProtocol)
    f3.d = Deferred()
    f3.d.addCallback(lambda data: eq.fire_eventually(data))

    d2 = guest.connector_for("hello").connect(f2)
    d3 = guest.connector_for("bonjour").connect(f3)
    await d2
    await d3

    data1 = await fserv0.d
    data2 = await f2.d
    assert data1 == b"hello\n"
    assert data2 == b"hello\n"

    data3 = await fserv1.d
    data4 = await f3.d
    assert data3 == data4 == b"bonjour\n"

    await w2.close()


class ReconP(Protocol):
    def eventually(self, which, data):
        d = self.factory.deferreds[which]
        self.factory.eq.fire_eventually(data).addCallback(d.callback)

    def connectionMade(self):
        self.eventually("connectionMade", self)
        # self.transport.write(b"hello\n")

    def dataReceived(self, data):
        self.eventually("dataReceived", data)

    def connectionLost(self, why):
        self.eventually("connectionLost", (self, why))


@implementer(IProtocolFactory)
class ReconF(Factory):
    protocol = ReconP
    subprotocol = "proto"

    def __init__(self, eq):
        Factory.__init__(self)
        self.eq = eq
        self.deferreds = {}
        for name in ["connectionMade", "dataReceived", "connectionLost"]:
            self.deferreds[name] = Deferred()

    def resetDeferred(self, name):
        d = Deferred()
        self.deferreds[name] = d
        return d


@implementer(IProtocolFactory)
class SubFac(Factory):
    pass


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_reconnect(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w2 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    await doBoth(w1.get_verifier(), w2.get_verifier())

    f1 = ReconF(eq)

    eps1 = w1.dilate()
    eps2 = w2.dilate()
    print("w.dilate ready")

    eps1.listener_for("proto").listen(f1)

    f2 = ReconF(eq)
    d2 = eps2.connector_for("proto").connect(f2)
    await d2

    protocols = {}

    def p_connected(p, index):
        protocols[index] = p
        msg = f"hello from {index}\n"
        p.transport.write(msg.encode("ascii"))
    f1.deferreds["connectionMade"].addCallback(p_connected, 1)
    f2.deferreds["connectionMade"].addCallback(p_connected, 2)

    data1 = await f1.deferreds["dataReceived"]
    data2 = await f2.deferreds["dataReceived"]
    assert data1 == b"hello from 2\n"
    assert data2 == b"hello from 1\n"
    # the ACKs are now in flight and may not arrive before we kill the
    # connection

    f1.resetDeferred("connectionMade")
    f2.resetDeferred("connectionMade")
    f1.resetDeferred("dataReceived")
    d2 = f2.resetDeferred("dataReceived")

    # now we reach inside and drop the connection
    sc = protocols[1].transport
    orig_connection = sc._manager._connection
    orig_connection.disconnect()

    # stall until the connection has been replaced
    await poll_until(lambda: sc._manager._connection
                     and (orig_connection != sc._manager._connection))

    # now write some more data, which should travel over the new
    # connection
    protocols[1].transport.write(b"more\n")
    data2 = await d2
    assert data2 == b"more\n"

    replacement_connection = sc._manager._connection
    assert orig_connection != replacement_connection

    # the application-visible Protocol should not observe the
    # interruption
    assert not f1.deferreds["connectionMade"].called
    assert not f2.deferreds["connectionMade"].called
    assert not f1.deferreds["connectionLost"].called
    assert not f2.deferreds["connectionLost"].called

    await w1.close()
    await w2.close()

@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_data_while_offline(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w2 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    await doBoth(w1.get_verifier(), w2.get_verifier())

    f1 = ReconF(eq)
    eps1 = w1.dilate()
    eps2 = w2.dilate()
    print("w.dilate ready")

    eps1.listener_for("proto").listen(f1)

    f2 = ReconF(eq)
    d2 = eps2.connector_for("proto").connect(f2)
    await d2

    protocols = {}

    def p_connected(p, index):
        protocols[index] = p
        msg = f"hello from {index}\n"
        p.transport.write(msg.encode("ascii"))
    f1.deferreds["connectionMade"].addCallback(p_connected, 1)
    f2.deferreds["connectionMade"].addCallback(p_connected, 2)

    data1 = await f1.deferreds["dataReceived"]
    data2 = await f2.deferreds["dataReceived"]
    assert data1 == b"hello from 2\n"
    assert data2 == b"hello from 1\n"
    # the ACKs are now in flight and may not arrive before we kill the
    # connection

    f1.resetDeferred("connectionMade")
    f2.resetDeferred("connectionMade")
    d1 = f1.resetDeferred("dataReceived")
    d2 = f2.resetDeferred("dataReceived")

    # switch off connections
    assert not w1._boss._D._manager._debug_stall_connector
    cd1, cd2 = Deferred(), Deferred()
    w1._boss._D._manager._debug_stall_connector = cd1.callback
    w2._boss._D._manager._debug_stall_connector = cd2.callback

    # now we reach inside and drop the connection
    sc = protocols[1].transport
    orig_connection = sc._manager._connection
    orig_connection.disconnect()

    c1 = await cd1
    c2 = await cd2
    assert IDilationConnector.providedBy(c1)
    assert IDilationConnector.providedBy(c2)
    assert c1 is not orig_connection
    w1._boss._D._manager._debug_stall_connector = False
    w2._boss._D._manager._debug_stall_connector = False

    # now write some data while the connection is definitely offline
    protocols[1].transport.write(b"more 1->2\n")
    protocols[2].transport.write(b"more 2->1\n")

    # allow the connections to proceed
    c1.start()
    c2.start()

    # and wait for the data to arrive
    data2 = await d2
    assert data2 == b"more 1->2\n"
    data1 = await d1
    assert data1 == b"more 2->1\n"

    # the application-visible Protocol should not observe the
    # interruption
    assert not f1.deferreds["connectionMade"].called
    assert not f2.deferreds["connectionMade"].called
    assert not f1.deferreds["connectionLost"].called
    assert not f2.deferreds["connectionLost"].called

    await w1.close()
    await w2.close()


@pytest_twisted.ensureDeferred()
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test_endpoints(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w2 = wormhole.create(APPID, mailbox.url, reactor, dilation=True)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    await doBoth(w1.get_verifier(), w2.get_verifier())

    f0 = ReconF(eq)
    eps1 = w1.dilate()
    eps2 = w2.dilate()
    print("w.dilate ready")
    eps1.listener_for("proto").listen(f0)

    from twisted.python import log
    f1 = ReconF(eq)
    log.msg("connecting")
    p1_client = await eps2.connector_for("proto").connect(f1)
    log.msg("sending c->s")
    p1_client.transport.write(b"hello from p1\n")
    data = await f0.deferreds["dataReceived"]
    assert data == b"hello from p1\n"
    p1_server = await f0.deferreds["connectionMade"]
    log.msg("sending s->c")
    p1_server.transport.write(b"hello p1\n")
    log.msg("waiting for client to receive")
    data = await f1.deferreds["dataReceived"]
    assert data == b"hello p1\n"

    # open a second channel
    f0.resetDeferred("connectionMade")
    f0.resetDeferred("dataReceived")
    f1.resetDeferred("dataReceived")
    f2 = ReconF(eq)
    p2_client = await eps2.connector_for("proto").connect(f2)
    p2_server = await f0.deferreds["connectionMade"]
    p2_server.transport.write(b"hello p2\n")
    data = await f2.deferreds["dataReceived"]
    assert data == b"hello p2\n"
    p2_client.transport.write(b"hello from p2\n")
    data = await f0.deferreds["dataReceived"]
    assert data == b"hello from p2\n"
    assert not f1.deferreds["dataReceived"].called

    # now close the first subchannel (p1) from the listener side
    p1_server.transport.loseConnection()
    await f0.deferreds["connectionLost"]
    await f1.deferreds["connectionLost"]

    f0.resetDeferred("connectionLost")
    # and close the second from the connector side
    p2_client.transport.loseConnection()
    await f0.deferreds["connectionLost"]
    await f2.deferreds["connectionLost"]

    await w1.close()
    await w2.close()



import os
from binascii import unhexlify

from zope.interface import implementer

from ._interfaces import IWordlist

# The PGP Word List, which maps bytes to phonetically-distinct words. There
# are two lists, even and odd, and encodings should alternate between them to
# detect dropped words. https://en.wikipedia.org/wiki/PGP_Words

# Thanks to Warren Guy for transcribing them:
# https://github.com/warrenguy/javascript-pgp-word-list


raw_words = {
    '00': ['aardvark', 'adroitness'],
    '01': ['absurd', 'adviser'],
    '02': ['accrue', 'aftermath'],
    '03': ['acme', 'aggregate'],
    '04': ['adrift', 'alkali'],
    '05': ['adult', 'almighty'],
    '06': ['afflict', 'amulet'],
    '07': ['ahead', 'amusement'],
    '08': ['aimless', 'antenna'],
    '09': ['Algol', 'applicant'],
    '0A': ['allow', 'Apollo'],
    '0B': ['alone', 'armistice'],
    '0C': ['ammo', 'article'],
    '0D': ['ancient', 'asteroid'],
    '0E': ['apple', 'Atlantic'],
    '0F': ['artist', 'atmosphere'],
    '10': ['assume', 'autopsy'],
    '11': ['Athens', 'Babylon'],
    '12': ['atlas', 'backwater'],
    '13': ['Aztec', 'barbecue'],
    '14': ['baboon', 'belowground'],
    '15': ['backfield', 'bifocals'],
    '16': ['backward', 'bodyguard'],
    '17': ['banjo', 'bookseller'],
    '18': ['beaming', 'borderline'],
    '19': ['bedlamp', 'bottomless'],
    '1A': ['beehive', 'Bradbury'],
    '1B': ['beeswax', 'bravado'],
    '1C': ['befriend', 'Brazilian'],
    '1D': ['Belfast', 'breakaway'],
    '1E': ['berserk', 'Burlington'],
    '1F': ['billiard', 'businessman'],
    '20': ['bison', 'butterfat'],
    '21': ['blackjack', 'Camelot'],
    '22': ['blockade', 'candidate'],
    '23': ['blowtorch', 'cannonball'],
    '24': ['bluebird', 'Capricorn'],
    '25': ['bombast', 'caravan'],
    '26': ['bookshelf', 'caretaker'],
    '27': ['brackish', 'celebrate'],
    '28': ['breadline', 'cellulose'],
    '29': ['breakup', 'certify'],
    '2A': ['brickyard', 'chambermaid'],
    '2B': ['briefcase', 'Cherokee'],
    '2C': ['Burbank', 'Chicago'],
    '2D': ['button', 'clergyman'],
    '2E': ['buzzard', 'coherence'],
    '2F': ['cement', 'combustion'],
    '30': ['chairlift', 'commando'],
    '31': ['chatter', 'company'],
    '32': ['checkup', 'component'],
    '33': ['chisel', 'concurrent'],
    '34': ['choking', 'confidence'],
    '35': ['chopper', 'conformist'],
    '36': ['Christmas', 'congregate'],
    '37': ['clamshell', 'consensus'],
    '38': ['classic', 'consulting'],
    '39': ['classroom', 'corporate'],
    '3A': ['cleanup', 'corrosion'],
    '3B': ['clockwork', 'councilman'],
    '3C': ['cobra', 'crossover'],
    '3D': ['commence', 'crucifix'],
    '3E': ['concert', 'cumbersome'],
    '3F': ['cowbell', 'customer'],
    '40': ['crackdown', 'Dakota'],
    '41': ['cranky', 'decadence'],
    '42': ['crowfoot', 'December'],
    '43': ['crucial', 'decimal'],
    '44': ['crumpled', 'designing'],
    '45': ['crusade', 'detector'],
    '46': ['cubic', 'detergent'],
    '47': ['dashboard', 'determine'],
    '48': ['deadbolt', 'dictator'],
    '49': ['deckhand', 'dinosaur'],
    '4A': ['dogsled', 'direction'],
    '4B': ['dragnet', 'disable'],
    '4C': ['drainage', 'disbelief'],
    '4D': ['dreadful', 'disruptive'],
    '4E': ['drifter', 'distortion'],
    '4F': ['dropper', 'document'],
    '50': ['drumbeat', 'embezzle'],
    '51': ['drunken', 'enchanting'],
    '52': ['Dupont', 'enrollment'],
    '53': ['dwelling', 'enterprise'],
    '54': ['eating', 'equation'],
    '55': ['edict', 'equipment'],
    '56': ['egghead', 'escapade'],
    '57': ['eightball', 'Eskimo'],
    '58': ['endorse', 'everyday'],
    '59': ['endow', 'examine'],
    '5A': ['enlist', 'existence'],
    '5B': ['erase', 'exodus'],
    '5C': ['escape', 'fascinate'],
    '5D': ['exceed', 'filament'],
    '5E': ['eyeglass', 'finicky'],
    '5F': ['eyetooth', 'forever'],
    '60': ['facial', 'fortitude'],
    '61': ['fallout', 'frequency'],
    '62': ['flagpole', 'gadgetry'],
    '63': ['flatfoot', 'Galveston'],
    '64': ['flytrap', 'getaway'],
    '65': ['fracture', 'glossary'],
    '66': ['framework', 'gossamer'],
    '67': ['freedom', 'graduate'],
    '68': ['frighten', 'gravity'],
    '69': ['gazelle', 'guitarist'],
    '6A': ['Geiger', 'hamburger'],
    '6B': ['glitter', 'Hamilton'],
    '6C': ['glucose', 'handiwork'],
    '6D': ['goggles', 'hazardous'],
    '6E': ['goldfish', 'headwaters'],
    '6F': ['gremlin', 'hemisphere'],
    '70': ['guidance', 'hesitate'],
    '71': ['hamlet', 'hideaway'],
    '72': ['highchair', 'holiness'],
    '73': ['hockey', 'hurricane'],
    '74': ['indoors', 'hydraulic'],
    '75': ['indulge', 'impartial'],
    '76': ['inverse', 'impetus'],
    '77': ['involve', 'inception'],
    '78': ['island', 'indigo'],
    '79': ['jawbone', 'inertia'],
    '7A': ['keyboard', 'infancy'],
    '7B': ['kickoff', 'inferno'],
    '7C': ['kiwi', 'informant'],
    '7D': ['klaxon', 'insincere'],
    '7E': ['locale', 'insurgent'],
    '7F': ['lockup', 'integrate'],
    '80': ['merit', 'intention'],
    '81': ['minnow', 'inventive'],
    '82': ['miser', 'Istanbul'],
    '83': ['Mohawk', 'Jamaica'],
    '84': ['mural', 'Jupiter'],
    '85': ['music', 'leprosy'],
    '86': ['necklace', 'letterhead'],
    '87': ['Neptune', 'liberty'],
    '88': ['newborn', 'maritime'],
    '89': ['nightbird', 'matchmaker'],
    '8A': ['Oakland', 'maverick'],
    '8B': ['obtuse', 'Medusa'],
    '8C': ['offload', 'megaton'],
    '8D': ['optic', 'microscope'],
    '8E': ['orca', 'microwave'],
    '8F': ['payday', 'midsummer'],
    '90': ['peachy', 'millionaire'],
    '91': ['pheasant', 'miracle'],
    '92': ['physique', 'misnomer'],
    '93': ['playhouse', 'molasses'],
    '94': ['Pluto', 'molecule'],
    '95': ['preclude', 'Montana'],
    '96': ['prefer', 'monument'],
    '97': ['preshrunk', 'mosquito'],
    '98': ['printer', 'narrative'],
    '99': ['prowler', 'nebula'],
    '9A': ['pupil', 'newsletter'],
    '9B': ['puppy', 'Norwegian'],
    '9C': ['python', 'October'],
    '9D': ['quadrant', 'Ohio'],
    '9E': ['quiver', 'onlooker'],
    '9F': ['quota', 'opulent'],
    'A0': ['ragtime', 'Orlando'],
    'A1': ['ratchet', 'outfielder'],
    'A2': ['rebirth', 'Pacific'],
    'A3': ['reform', 'pandemic'],
    'A4': ['regain', 'Pandora'],
    'A5': ['reindeer', 'paperweight'],
    'A6': ['rematch', 'paragon'],
    'A7': ['repay', 'paragraph'],
    'A8': ['retouch', 'paramount'],
    'A9': ['revenge', 'passenger'],
    'AA': ['reward', 'pedigree'],
    'AB': ['rhythm', 'Pegasus'],
    'AC': ['ribcage', 'penetrate'],
    'AD': ['ringbolt', 'perceptive'],
    'AE': ['robust', 'performance'],
    'AF': ['rocker', 'pharmacy'],
    'B0': ['ruffled', 'phonetic'],
    'B1': ['sailboat', 'photograph'],
    'B2': ['sawdust', 'pioneer'],
    'B3': ['scallion', 'pocketful'],
    'B4': ['scenic', 'politeness'],
    'B5': ['scorecard', 'positive'],
    'B6': ['Scotland', 'potato'],
    'B7': ['seabird', 'processor'],
    'B8': ['select', 'provincial'],
    'B9': ['sentence', 'proximate'],
    'BA': ['shadow', 'puberty'],
    'BB': ['shamrock', 'publisher'],
    'BC': ['showgirl', 'pyramid'],
    'BD': ['skullcap', 'quantity'],
    'BE': ['skydive', 'racketeer'],
    'BF': ['slingshot', 'rebellion'],
    'C0': ['slowdown', 'recipe'],
    'C1': ['snapline', 'recover'],
    'C2': ['snapshot', 'repellent'],
    'C3': ['snowcap', 'replica'],
    'C4': ['snowslide', 'reproduce'],
    'C5': ['solo', 'resistor'],
    'C6': ['southward', 'responsive'],
    'C7': ['soybean', 'retraction'],
    'C8': ['spaniel', 'retrieval'],
    'C9': ['spearhead', 'retrospect'],
    'CA': ['spellbind', 'revenue'],
    'CB': ['spheroid', 'revival'],
    'CC': ['spigot', 'revolver'],
    'CD': ['spindle', 'sandalwood'],
    'CE': ['spyglass', 'sardonic'],
    'CF': ['stagehand', 'Saturday'],
    'D0': ['stagnate', 'savagery'],
    'D1': ['stairway', 'scavenger'],
    'D2': ['standard', 'sensation'],
    'D3': ['stapler', 'sociable'],
    'D4': ['steamship', 'souvenir'],
    'D5': ['sterling', 'specialist'],
    'D6': ['stockman', 'speculate'],
    'D7': ['stopwatch', 'stethoscope'],
    'D8': ['stormy', 'stupendous'],
    'D9': ['sugar', 'supportive'],
    'DA': ['surmount', 'surrender'],
    'DB': ['suspense', 'suspicious'],
    'DC': ['sweatband', 'sympathy'],
    'DD': ['swelter', 'tambourine'],
    'DE': ['tactics', 'telephone'],
    'DF': ['talon', 'therapist'],
    'E0': ['tapeworm', 'tobacco'],
    'E1': ['tempest', 'tolerance'],
    'E2': ['tiger', 'tomorrow'],
    'E3': ['tissue', 'torpedo'],
    'E4': ['tonic', 'tradition'],
    'E5': ['topmost', 'travesty'],
    'E6': ['tracker', 'trombonist'],
    'E7': ['transit', 'truncated'],
    'E8': ['trauma', 'typewriter'],
    'E9': ['treadmill', 'ultimate'],
    'EA': ['Trojan', 'undaunted'],
    'EB': ['trouble', 'underfoot'],
    'EC': ['tumor', 'unicorn'],
    'ED': ['tunnel', 'unify'],
    'EE': ['tycoon', 'universe'],
    'EF': ['uncut', 'unravel'],
    'F0': ['unearth', 'upcoming'],
    'F1': ['unwind', 'vacancy'],
    'F2': ['uproot', 'vagabond'],
    'F3': ['upset', 'vertigo'],
    'F4': ['upshot', 'Virginia'],
    'F5': ['vapor', 'visitor'],
    'F6': ['village', 'vocalist'],
    'F7': ['virus', 'voyager'],
    'F8': ['Vulcan', 'warranty'],
    'F9': ['waffle', 'Waterloo'],
    'FA': ['wallet', 'whimsical'],
    'FB': ['watchword', 'Wichita'],
    'FC': ['wayside', 'Wilmington'],
    'FD': ['willow', 'Wyoming'],
    'FE': ['woodlark', 'yesteryear'],
    'FF': ['Zulu', 'Yucatan']
}

byte_to_even_word = {unhexlify(k.encode("ascii")): both_words[0]
                          for k, both_words in raw_words.items()}

byte_to_odd_word = {unhexlify(k.encode("ascii")): both_words[1]
                         for k, both_words in raw_words.items()}

even_words_lowercase, odd_words_lowercase = set(), set()

for k, both_words in raw_words.items():
    even_word, odd_word = both_words
    even_words_lowercase.add(even_word.lower())
    odd_words_lowercase.add(odd_word.lower())


@implementer(IWordlist)
class PGPWordList:
    def get_completions(self, prefix, num_words=2):
        # start with the odd words
        count = prefix.count("-")
        if count % 2 == 0:
            words = odd_words_lowercase
        else:
            words = even_words_lowercase
        last_partial_word = prefix.split("-")[-1]
        lp = len(last_partial_word)
        completions = set()
        for word in words:
            if word.startswith(last_partial_word):
                if lp == 0:
                    suffix = prefix + word
                else:
                    suffix = prefix[:-lp] + word
                # append a hyphen if we expect more words
                if count + 1 < num_words:
                    suffix += "-"
                completions.add(suffix)
        return completions

    def choose_words(self, length):
        words = []
        for i in range(length):
            # we start with an "odd word"
            if i % 2 == 0:
                words.append(byte_to_odd_word[os.urandom(1)].lower())
            else:
                words.append(byte_to_even_word[os.urandom(1)].lower())
        return "-".join(words)



import sys
import re
from collections import namedtuple
from twisted.internet.endpoints import TCP4ClientEndpoint, TCP6ClientEndpoint, HostnameEndpoint
from twisted.internet.abstract import isIPAddress, isIPv6Address
from twisted.python import log

# These namedtuples are "hint objects". The JSON-serializable dictionaries
# are "hint dicts".

# DirectTCPV1Hint and TorTCPV1Hint mean the following protocol:
# * make a TCP connection (possibly via Tor)
# * send the sender/receiver handshake bytes first
# * expect to see the receiver/sender handshake bytes from the other side
# * the sender writes "go\n", the receiver waits for "go\n"
# * the rest of the connection contains transit data
DirectTCPV1Hint = namedtuple("DirectTCPV1Hint",
                             ["hostname", "port", "priority"])
TorTCPV1Hint = namedtuple("TorTCPV1Hint", ["hostname", "port", "priority"])
# RelayV1Hint contains a tuple of DirectTCPV1Hint and TorTCPV1Hint hints (we
# use a tuple rather than a list so they'll be hashable into a set). For each
# one, make the TCP connection, send the relay handshake, then complete the
# rest of the V1 protocol. Only one hint per relay is useful.
RelayV1Hint = namedtuple("RelayV1Hint", ["hints"])


def describe_hint_obj(hint, relay, tor):
    prefix = "tor->" if tor else "->"
    if relay:
        prefix = prefix + "relay:"
    if isinstance(hint, DirectTCPV1Hint):
        return prefix + "tcp:%s:%d" % (hint.hostname, hint.port)
    elif isinstance(hint, TorTCPV1Hint):
        return prefix + "tor:%s:%d" % (hint.hostname, hint.port)
    else:
        return prefix + str(hint)


def parse_hint_argv(hint, stderr=sys.stderr):
    assert isinstance(hint, str)
    # return tuple or None for an unparseable hint
    priority = 0.0
    # parse hint type
    mo = re.search(r'^([a-zA-Z0-9]+):(.*)$', hint)
    if not mo:
        print(f"unparseable hint '{hint}'", file=stderr)
        return None
    hint_type = mo.group(1)
    if hint_type != "tcp":
        print(f"unknown hint type '{hint_type}' in '{hint}'",
              file=stderr)
        return None
    hint_value = mo.group(2)
    hint_host = ""
    pieces = []  # hint_value split into pieces
    # parse IPv6 address (must have square brackets)
    mo = re.search(r'^\[([a-zA-Z0-9:]+)\]:(.*)$', hint_value)
    if mo:
        hint_host = mo.group(1)
        if not isIPv6Address(hint_host):
            print(f"invalid IPv6 address in TCP hint '{hint}'",
                  file=stderr)
            return None
        pieces = [hint_host] + mo.group(2).split(":")
    # if not IPv6, parse IPv4 address or hostname
    else:
        pieces = hint_value.split(":")
        if len(pieces) < 2:
            print(f"unparseable TCP hint (need more colons) '{hint}'",
                  file=stderr)
            return None
        hint_host = pieces[0]
    # parse the port:
    mo = re.search(r'^(\d+)$', pieces[1])
    if not mo:
        print(f"non-numeric port in TCP hint '{hint}'", file=stderr)
        return None
    hint_port = int(pieces[1])
    # parse the rest ("priority=float")
    for more in pieces[2:]:
        if more.startswith("priority="):
            more_pieces = more.split("=")
            try:
                priority = float(more_pieces[1])
            except ValueError:
                print(f"non-float priority= in TCP hint '{hint}'",
                      file=stderr)
                return None
    return DirectTCPV1Hint(hint_host, hint_port, priority)


def endpoint_from_hint_obj(hint, tor, reactor):
    if tor:
        if isinstance(hint, (DirectTCPV1Hint, TorTCPV1Hint)):
            # this Tor object will throw ValueError for non-public IPv4
            # addresses and any IPv6 address
            try:
                return tor.stream_via(hint.hostname, hint.port)
            except ValueError:
                return None
        return None
    if isinstance(hint, DirectTCPV1Hint):
        # avoid DNS lookup unless necessary
        if isIPAddress(hint.hostname):
            return TCP4ClientEndpoint(reactor, hint.hostname, hint.port)
        if isIPv6Address(hint.hostname):
            return TCP6ClientEndpoint(reactor, hint.hostname, hint.port)
        return HostnameEndpoint(reactor, hint.hostname, hint.port)
    return None


def parse_tcp_v1_hint(hint):  # hint_struct -> hint_obj
    hint_type = hint.get("type", "")
    if hint_type not in ["direct-tcp-v1", "tor-tcp-v1"]:
        log.msg(f"unknown hint type: {hint!r}")
        return None
    if not ("hostname" in hint and
            isinstance(hint["hostname"], str)):
        log.msg(f"invalid hostname in hint: {hint!r}")
        return None
    if not ("port" in hint and
            isinstance(hint["port"], int)):
        log.msg(f"invalid port in hint: {hint!r}")
        return None
    priority = hint.get("priority", 0.0)
    if hint_type == "direct-tcp-v1":
        return DirectTCPV1Hint(hint["hostname"], hint["port"], priority)
    else:
        return TorTCPV1Hint(hint["hostname"], hint["port"], priority)


def parse_hint(hint_struct):
    hint_type = hint_struct.get("type", "")
    if hint_type == "relay-v1":
        # the struct can include multiple ways to reach the same relay
        rhints = filter(lambda h: h,  # drop None (unrecognized)
                        [parse_tcp_v1_hint(rh) for rh in hint_struct["hints"]])
        return RelayV1Hint(list(rhints))
    return parse_tcp_v1_hint(hint_struct)


def encode_hint(h):
    if isinstance(h, DirectTCPV1Hint):
        return {"type": "direct-tcp-v1",
                "priority": h.priority,
                "hostname": h.hostname,
                "port": h.port,  # integer
                }
    elif isinstance(h, RelayV1Hint):
        rhint = {"type": "relay-v1", "hints": []}
        for rh in h.hints:
            rhint["hints"].append({"type": "direct-tcp-v1",
                                   "priority": rh.priority,
                                   "hostname": rh.hostname,
                                   "port": rh.port})
        return rhint
    elif isinstance(h, TorTCPV1Hint):
        return {"type": "tor-tcp-v1",
                "priority": h.priority,
                "hostname": h.hostname,
                "port": h.port,  # integer
                }
    raise ValueError("unknown hint type", h)



from .. import journal
from .._interfaces import IJournal

def test_journal():
    events = []
    j = journal.Journal(lambda: events.append("checkpoint"))
    assert (IJournal.providedBy(j))

    with j.process():
        j.queue_outbound(events.append, "message1")
        j.queue_outbound(events.append, "message2")
        assert events == [],"test_journal: events list is not empty"
    assert events == ["checkpoint", "message1", "message2"]

def test_immediate():
    events = []
    j = journal.ImmediateJournal()
    assert IJournal.providedBy(j)

    with j.process():
        j.queue_outbound(events.append, "message1")
        assert events == ["message1"]
        j.queue_outbound(events.append, "message2")
        assert events == ["message1", "message2"]
    assert events == ["message1", "message2"]



from ._rlcompleter import input_with_completion
from .wormhole import create
from ._status import WormholeStatus, DilationStatus  # export as public API
from ._dilation.subchannel import SubchannelAddress

from . import _version
__version__ = _version.get_versions()['version']

__all__ = [
    "__version__",
    "create", "input_with_completion",
    "WormholeStatus",

    # Dilation-related exports
    "DilationStatus",
    "SubchannelAddress",
]



from hashlib import sha256

from attr import attrib, attrs
from attr.validators import instance_of
from automat import MethodicalMachine
from nacl import utils
from nacl.exceptions import CryptoError
from nacl.secret import SecretBox
from spake2 import SPAKE2_Symmetric
from zope.interface import implementer

from . import _interfaces
from .util import (bytes_to_dict, bytes_to_hexstr, dict_to_bytes,
                   hexstr_to_bytes, to_bytes, HKDF, provides)

CryptoError
__all__ = ["derive_key", "derive_phase_key", "CryptoError", "Key"]


def derive_key(key, purpose, length=SecretBox.KEY_SIZE):
    if not isinstance(key, bytes):
        raise TypeError(type(key))
    if not isinstance(purpose, bytes):
        raise TypeError(type(purpose))
    if not isinstance(length, int):
        raise TypeError(type(length))
    return HKDF(key, length, CTXinfo=purpose)


def derive_phase_key(key, side, phase):
    assert isinstance(side, str), type(side)
    assert isinstance(phase, str), type(phase)
    side_bytes = side.encode("ascii")
    phase_bytes = phase.encode("ascii")
    purpose = (b"wormhole:phase:" + sha256(side_bytes).digest() +
               sha256(phase_bytes).digest())
    return derive_key(key, purpose)


def decrypt_data(key, encrypted):
    assert isinstance(key, bytes), type(key)
    assert isinstance(encrypted, bytes), type(encrypted)
    assert len(key) == SecretBox.KEY_SIZE, len(key)
    box = SecretBox(key)
    data = box.decrypt(encrypted)
    return data


def encrypt_data(key, plaintext):
    assert isinstance(key, bytes), type(key)
    assert isinstance(plaintext, bytes), type(plaintext)
    assert len(key) == SecretBox.KEY_SIZE, len(key)
    box = SecretBox(key)
    nonce = utils.random(SecretBox.NONCE_SIZE)
    return box.encrypt(plaintext, nonce)


# the Key we expose to callers (Boss, Ordering) is responsible for sorting
# the two messages (got_code and got_pake), then delivering them to
# _SortedKey in the right order.


@attrs
@implementer(_interfaces.IKey)
class Key:
    _appid = attrib(validator=instance_of(str))
    _versions = attrib(validator=instance_of(dict))
    _side = attrib(validator=instance_of(str))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._SK = _SortedKey(self._appid, self._versions, self._side,
                              self._timing)
        self._debug_pake_stashed = False  # for tests

    def wire(self, boss, mailbox, receive):
        self._SK.wire(boss, mailbox, receive)

    @m.state(initial=True)
    def S00(self):
        pass  # pragma: no cover

    @m.state()
    def S01(self):
        pass  # pragma: no cover

    @m.state()
    def S10(self):
        pass  # pragma: no cover

    @m.state()
    def S11(self):
        pass  # pragma: no cover

    @m.input()
    def got_code(self, code):
        pass

    @m.input()
    def got_pake(self, body):
        pass

    @m.output()
    def stash_pake(self, body):
        self._pake = body
        self._debug_pake_stashed = True

    @m.output()
    def deliver_code(self, code):
        self._SK.got_code(code)

    @m.output()
    def deliver_pake(self, body):
        self._SK.got_pake(body)

    @m.output()
    def deliver_code_and_stashed_pake(self, code):
        self._SK.got_code(code)
        self._SK.got_pake(self._pake)

    S00.upon(got_code, enter=S10, outputs=[deliver_code])
    S10.upon(got_pake, enter=S11, outputs=[deliver_pake])
    S00.upon(got_pake, enter=S01, outputs=[stash_pake])
    S01.upon(got_code, enter=S11, outputs=[deliver_code_and_stashed_pake])


@attrs
class _SortedKey:
    _appid = attrib(validator=instance_of(str))
    _versions = attrib(validator=instance_of(dict))
    _side = attrib(validator=instance_of(str))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def wire(self, boss, mailbox, receive):
        self._B = _interfaces.IBoss(boss)
        self._M = _interfaces.IMailbox(mailbox)
        self._R = _interfaces.IReceive(receive)

    @m.state(initial=True)
    def S0_know_nothing(self):
        pass  # pragma: no cover

    @m.state()
    def S1_know_code(self):
        pass  # pragma: no cover

    @m.state()
    def S2_know_key(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S3_scared(self):
        pass  # pragma: no cover

    # from Boss
    @m.input()
    def got_code(self, code):
        pass

    # from Ordering
    def got_pake(self, body):
        assert isinstance(body, bytes), type(body)
        payload = bytes_to_dict(body)
        if "pake_v1" in payload:
            self.got_pake_good(hexstr_to_bytes(payload["pake_v1"]))
        else:
            self.got_pake_bad()

    @m.input()
    def got_pake_good(self, msg2):
        pass

    @m.input()
    def got_pake_bad(self):
        pass

    @m.output()
    def build_pake(self, code):
        with self._timing.add("pake1", waiting="crypto"):
            self._sp = SPAKE2_Symmetric(
                to_bytes(code), idSymmetric=to_bytes(self._appid))
            msg1 = self._sp.start()
        body = dict_to_bytes({"pake_v1": bytes_to_hexstr(msg1)})
        self._M.add_message("pake", body)

    @m.output()
    def scared(self):
        self._B.scared()

    @m.output()
    def compute_key(self, msg2):
        assert isinstance(msg2, bytes)
        with self._timing.add("pake2", waiting="crypto"):
            key = self._sp.finish(msg2)
        # TODO: make B.got_key() an eventual send, since it will fire the
        # user/application-layer get_unverified_key() Deferred, and if that
        # calls back into other wormhole APIs, bad things will happen
        self._B.got_key(key)
        phase = "version"
        data_key = derive_phase_key(key, self._side, phase)
        plaintext = dict_to_bytes(self._versions)
        encrypted = encrypt_data(data_key, plaintext)
        self._M.add_message(phase, encrypted)
        # TODO: R.got_key() needs to be eventual-send too, as it can trigger
        # app-level got_verifier() and got_message() Deferreds.
        self._R.got_key(key)

    S0_know_nothing.upon(got_code, enter=S1_know_code, outputs=[build_pake])
    S1_know_code.upon(got_pake_good, enter=S2_know_key, outputs=[compute_key])
    S1_know_code.upon(got_pake_bad, enter=S3_scared, outputs=[scared])



class WormholeError(Exception):
    """Parent class for all wormhole-related errors"""


class UnsendableFileError(Exception):
    """
    A file you wanted to send couldn't be read, maybe because it's not
    a file, or because it's a symlink that points to something
    that doesn't exist.

    To ignore this kind of error, you can run wormhole with the
    --ignore-unsendable-files flag.
    """


class ServerError(WormholeError):
    """The relay server complained about something we did."""


class ServerConnectionError(WormholeError):
    """We had a problem connecting to the relay server:"""

    def __init__(self, url, reason):
        self.url = url
        self.reason = reason

    def __str__(self):
        return str(self.reason)


class Timeout(WormholeError):
    pass


class WelcomeError(WormholeError):
    """
    The relay server told us to signal an error, probably because our version
    is too old to possibly work. The server said:"""
    pass


class LonelyError(WormholeError):
    """wormhole.close() was called before the peer connection could be
    established"""


class WrongPasswordError(WormholeError):
    """
    Key confirmation failed. Either you or your correspondent typed the code
    wrong, or a would-be man-in-the-middle attacker guessed incorrectly. Try
    sending the file again.
    """
    # or the data blob was corrupted, and that's why decrypt failed
    pass


class KeyFormatError(WormholeError):
    """
    The key you entered contains spaces or was missing a dash. Magic-wormhole
    expects the numerical nameplate and the code words to be separated by
    dashes. Please reenter the key you were given separating the words with
    dashes.
    """


class ReflectionAttack(WormholeError):
    """An attacker (or bug) reflected our outgoing message back to us."""


class InternalError(WormholeError):
    """The programmer did something wrong."""


class TransferError(WormholeError):
    """Something bad happened and the transfer failed."""


class NoTorError(WormholeError):
    """--tor was requested, but 'txtorcon' is not installed."""


class NoKeyError(WormholeError):
    """w.derive_key() was called before got_verifier() fired"""


class OnlyOneCodeError(WormholeError):
    """Only one w.generate_code/w.set_code/w.input_code may be called"""


class MustChooseNameplateFirstError(WormholeError):
    """The InputHelper was asked to do get_word_completions() or
    choose_words() before the nameplate was chosen."""


class AlreadyChoseNameplateError(WormholeError):
    """The InputHelper was asked to do get_nameplate_completions() after
    choose_nameplate() was called, or choose_nameplate() was called a second
    time."""


class AlreadyChoseWordsError(WormholeError):
    """The InputHelper was asked to do get_word_completions() after
    choose_words() was called, or choose_words() was called a second time."""


class AlreadyInputNameplateError(WormholeError):
    """The CodeInputter was asked to do completion on a nameplate, when we
    had already committed to a different one."""


class WormholeClosed(Exception):
    """Deferred-returning API calls errback with WormholeClosed if the
    wormhole was already closed, or if it closes before a real result can be
    obtained."""


class _UnknownPhaseError(Exception):
    """internal exception type, for tests."""


class _UnknownMessageTypeError(Exception):
    """internal exception type, for tests."""



from unittest import mock
from zope.interface import alsoProvides
from twisted.internet.interfaces import ITransport
from ..._dilation.connection import _Framer, Frame, Prologue, Disconnect
import pytest


def make_framer():
    t = mock.Mock()
    alsoProvides(t, ITransport)
    f = _Framer(t, b"outbound_prologue\n", b"inbound_prologue\n")
    return f, t


def test_bad_prologue_length():
    f, t = make_framer()
    assert t.mock_calls == []

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    t.mock_calls[:] = []
    assert [] == list(f.add_and_parse(b"inbound_"))  # wait for it
    assert t.mock_calls == []

    with mock.patch("wormhole._dilation.connection.log.msg") as m:
        with pytest.raises(Disconnect):
            list(f.add_and_parse(b"not the prologue after all"))
    assert m.mock_calls == \
                     [mock.call(f"bad prologue: {b'inbound_not the p'}")]
    assert t.mock_calls == []


def test_bad_prologue_newline():
    f, t = make_framer()
    assert t.mock_calls == []

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    t.mock_calls[:] = []
    assert [] == list(f.add_and_parse(b"inbound_"))  # wait for it

    assert [] == list(f.add_and_parse(b"not"))
    with mock.patch("wormhole._dilation.connection.log.msg") as m:
        with pytest.raises(Disconnect):
            list(f.add_and_parse(b"\n"))
    assert m.mock_calls == \
                     [mock.call("bad prologue: {}".format(
                         b"inbound_not\n"))]
    assert t.mock_calls == []


def test_good_prologue():
    f, t = make_framer()
    assert t.mock_calls == []

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    t.mock_calls[:] = []
    assert [Prologue()] == \
                     list(f.add_and_parse(b"inbound_prologue\n"))
    assert t.mock_calls == []

    # now send_frame should work
    f.send_frame(b"frame")
    assert t.mock_calls == \
                     [mock.call.write(b"\x00\x00\x00\x05frame")]


def test_bad_relay():
    f, t = make_framer()
    assert t.mock_calls == []
    f.use_relay(b"relay handshake\n")

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"relay handshake\n")]
    t.mock_calls[:] = []
    with mock.patch("wormhole._dilation.connection.log.msg") as m:
        with pytest.raises(Disconnect):
            list(f.add_and_parse(b"goodbye\n"))
    assert m.mock_calls == \
                     [mock.call(f"bad relay_ok: {b'goo'}")]
    assert t.mock_calls == []


def test_good_relay():
    f, t = make_framer()
    assert t.mock_calls == []
    f.use_relay(b"relay handshake\n")
    assert t.mock_calls == []

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"relay handshake\n")]
    t.mock_calls[:] = []

    assert [] == list(f.add_and_parse(b"ok\n"))
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]


def test_frame():
    f, t = make_framer()
    assert t.mock_calls == []

    f.connectionMade()
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    t.mock_calls[:] = []
    assert [Prologue()] == \
                     list(f.add_and_parse(b"inbound_prologue\n"))
    assert t.mock_calls == []

    encoded_frame = b"\x00\x00\x00\x05frame"
    assert [] == list(f.add_and_parse(encoded_frame[:2]))
    assert [] == list(f.add_and_parse(encoded_frame[2:6]))
    assert [Frame(frame=b"frame")] == \
                     list(f.add_and_parse(encoded_frame[6:]))



from unittest import mock

from .._wordlist import PGPWordList

def test_completions():
    wl = PGPWordList()
    gc = wl.get_completions
    assert gc("ar", 2) == {"armistice-", "article-"}
    assert gc("armis", 2) == {"armistice-"}
    assert gc("armistice", 2) == {"armistice-"}
    lots = gc("armistice-", 2)
    assert len(lots) == 256
    first = list(lots)[0]
    assert first.startswith("armistice-")
    assert gc("armistice-ba", 2) == { "armistice-baboon", "armistice-backfield",
            "armistice-backward", "armistice-banjo" }
    assert gc("armistice-ba", 3) == { "armistice-baboon-", "armistice-backfield-",
        "armistice-backward-", "armistice-banjo-" }
    assert gc("armistice-baboon", 2) == {"armistice-baboon"}
    assert gc("armistice-baboon", 3) == {"armistice-baboon-"}
    assert gc("armistice-baboon", 4) == {"armistice-baboon-"}


def test_choose_words():
    wl = PGPWordList()
    with mock.patch("os.urandom", side_effect=[b"\x04", b"\x10"]):
        assert wl.choose_words(2) == "alkali-assume"



import json
import time

from zope.interface import implementer

from ._interfaces import ITiming


class Event:
    def __init__(self, name, when, **details):
        # data fields that will be dumped to JSON later
        self._name = name
        self._start = time.time() if when is None else float(when)
        self._stop = None
        self._details = details

    def detail(self, **details):
        self._details.update(details)

    def finish(self, when=None, **details):
        self._stop = time.time() if when is None else float(when)
        self.detail(**details)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.finish()


@implementer(ITiming)
class DebugTiming:
    def __init__(self):
        self._events = []

    def add(self, name, when=None, **details):
        ev = Event(name, when, **details)
        self._events.append(ev)
        return ev

    def write(self, fn, stderr):
        with open(fn, "w") as f:
            data = [
                dict(
                    name=e._name,
                    start=e._start,
                    stop=e._stop,
                    details=e._details,
                ) for e in self._events
            ]
            json.dump(data, f, indent=1)
            f.write("\n")
        print(f"Timing data written to {fn}", file=stderr)



import errno
import os
import re
import subprocess
from unittest import mock

from .. import ipaddrs

DOTTED_QUAD_RE = re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$")

MOCK_IPADDR_OUTPUT = """\
1: lo: <LOOPBACK,UP,LOWER_UP> mtu 16436 qdisc noqueue state UNKNOWN \n\
    link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
    inet 127.0.0.1/8 scope host lo
    inet6 ::1/128 scope host \n\
       valid_lft forever preferred_lft forever
2: eth1: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc pfifo_fast state UP \
qlen 1000
    link/ether d4:3d:7e:01:b4:3e brd ff:ff:ff:ff:ff:ff
    inet 192.168.0.6/24 brd 192.168.0.255 scope global eth1
    inet6 fe80::d63d:7eff:fe01:b43e/64 scope link \n\
       valid_lft forever preferred_lft forever
3: wlan0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc mq state UP qlen\
 1000
    link/ether 90:f6:52:27:15:0a brd ff:ff:ff:ff:ff:ff
    inet 192.168.0.2/24 brd 192.168.0.255 scope global wlan0
    inet6 fe80::92f6:52ff:fe27:150a/64 scope link \n\
       valid_lft forever preferred_lft forever
"""

MOCK_IFCONFIG_OUTPUT = """\
eth1      Link encap:Ethernet  HWaddr d4:3d:7e:01:b4:3e  \n\
          inet addr:192.168.0.6  Bcast:192.168.0.255  Mask:255.255.255.0
          inet6 addr: fe80::d63d:7eff:fe01:b43e/64 Scope:Link
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:154242234 errors:0 dropped:0 overruns:0 frame:0
          TX packets:155461891 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000 \n\
          RX bytes:84367213640 (78.5 GiB)  TX bytes:73401695329 (68.3 GiB)
          Interrupt:20 Memory:f4f00000-f4f20000 \n\

lo        Link encap:Local Loopback  \n\
          inet addr:127.0.0.1  Mask:255.0.0.0
          inet6 addr: ::1/128 Scope:Host
          UP LOOPBACK RUNNING  MTU:16436  Metric:1
          RX packets:27449267 errors:0 dropped:0 overruns:0 frame:0
          TX packets:27449267 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:0 \n\
          RX bytes:192643017823 (179.4 GiB)  TX bytes:192643017823 (179.4 GiB)

wlan0     Link encap:Ethernet  HWaddr 90:f6:52:27:15:0a  \n\
          inet addr:192.168.0.2  Bcast:192.168.0.255  Mask:255.255.255.0
          inet6 addr: fe80::92f6:52ff:fe27:150a/64 Scope:Link
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:12352750 errors:0 dropped:0 overruns:0 frame:0
          TX packets:4501451 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000 \n\
          RX bytes:3916475942 (3.6 GiB)  TX bytes:458353654 (437.1 MiB)
"""

# This is actually from a VirtualBox VM running XP.
MOCK_ROUTE_OUTPUT = """\
===========================================================================
Interface List
0x1 ........................... MS TCP Loopback interface
0x2 ...08 00 27 c3 80 ad ...... AMD PCNET Family PCI Ethernet Adapter - \
Packet Scheduler Miniport
===========================================================================
===========================================================================
Active Routes:
Network Destination        Netmask          Gateway       Interface  Metric
          0.0.0.0          0.0.0.0         10.0.2.2       10.0.2.15       20
         10.0.2.0    255.255.255.0        10.0.2.15       10.0.2.15       20
        10.0.2.15  255.255.255.255        127.0.0.1       127.0.0.1       20
   10.255.255.255  255.255.255.255        10.0.2.15       10.0.2.15       20
        127.0.0.0        255.0.0.0        127.0.0.1       127.0.0.1       1
        224.0.0.0        240.0.0.0        10.0.2.15       10.0.2.15       20
  255.255.255.255  255.255.255.255        10.0.2.15       10.0.2.15       1
Default Gateway:          10.0.2.2
===========================================================================
Persistent Routes:
  None
"""

UNIX_TEST_ADDRESSES = {"127.0.0.1", "192.168.0.6", "192.168.0.2"}
WINDOWS_TEST_ADDRESSES = {"127.0.0.1", "10.0.2.15"}
CYGWIN_TEST_ADDRESSES = {"127.0.0.1"}


class FakeProcess:
    def __init__(self, output, err):
        self.output = output
        self.err = err

    def communicate(self):
        return (self.output, self.err)


def test_list():
    addresses = ipaddrs.find_addresses()
    assert "127.0.0.1" in addresses
    assert "0.0.0.0" not in addresses

# David A.'s OpenSolaris box timed out on this test one time when it was at
# 2s.
test_list.timeout = 4

def _test_list_mock(command, output, expected):
    first = True

    def call_Popen(args,
                   bufsize=0,
                   executable=None,
                   stdin=None,
                   stdout=None,
                   stderr=None,
                   preexec_fn=None,
                   close_fds=False,
                   shell=False,
                   cwd=None,
                   env=None,
                   universal_newlines=False,
                   startupinfo=None,
                   creationflags=0):
        nonlocal first
        if first:
            first = False
            e = OSError("EINTR")
            e.errno = errno.EINTR
            raise e
        elif os.path.basename(args[0]) == command:
            return FakeProcess(output, "")
        else:
            e = OSError("[Errno 2] No such file or directory")
            e.errno = errno.ENOENT
            raise e

    def call_which(name):
        return [name]

    patch_popen = mock.patch.object(subprocess, 'Popen', call_Popen)
    patch_isfile = mock.patch.object(os.path, 'isfile', lambda x: True)
    patch_which = mock.patch.object(ipaddrs, 'which', call_which)

    with patch_popen, patch_isfile, patch_which:
        addresses = ipaddrs.find_addresses()

    assert set(addresses) == set(expected)

def test_list_mock_ip_addr():
    with mock.patch.object(ipaddrs, 'platform', "linux2"):
        _test_list_mock("ip", MOCK_IPADDR_OUTPUT, UNIX_TEST_ADDRESSES)

def test_list_mock_ifconfig():
    with mock.patch.object(ipaddrs, 'platform', "linux2"):
        _test_list_mock("ifconfig", MOCK_IFCONFIG_OUTPUT,
                        UNIX_TEST_ADDRESSES)

def test_list_mock_route():
    with mock.patch.object(ipaddrs, 'platform', "win32"):
        _test_list_mock("route.exe", MOCK_ROUTE_OUTPUT,
                        WINDOWS_TEST_ADDRESSES)

def test_list_mock_cygwin():
    with mock.patch.object(ipaddrs, 'platform', "cygwin"):
        _test_list_mock(None, None, CYGWIN_TEST_ADDRESSES)



from collections import namedtuple
from itertools import cycle
from unittest import mock
from zope.interface import alsoProvides
from twisted.internet.task import Clock, Cooperator
from twisted.internet.interfaces import IPullProducer
from ...eventual import EventualQueue
from ..._interfaces import IDilationManager
from ..._dilation.connection import KCM, Open, Data, Close, Ack
from ..._dilation.outbound import Outbound, PullToPush
from .common import clear_mock_calls
import pytest

Pauser = namedtuple("Pauser", ["seqnum"])
NonPauser = namedtuple("NonPauser", ["seqnum"])
Stopper = namedtuple("Stopper", ["sc"])


def make_outbound():
    m = mock.Mock()
    alsoProvides(m, IDilationManager)
    clock = Clock()
    eq = EventualQueue(clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    coop = Cooperator(terminationPredicateFactory=term_factory,
                      scheduler=eq.eventually)
    o = Outbound(m, coop)
    c = mock.Mock()  # Connection

    def maybe_pause(r):
        if isinstance(r, Pauser):
            o.pauseProducing()
        elif isinstance(r, Stopper):
            o.subchannel_unregisterProducer(r.sc)
    c.send_record = mock.Mock(side_effect=maybe_pause)
    o._test_eq = eq
    o._test_term = term
    return o, m, c


def test_build_record():
    o, m, c = make_outbound()
    scid1 = b"scid"
    assert o.build_record(Open, scid1, b"proto") == \
                     Open(seqnum=0, scid=b"scid", subprotocol=b"proto")
    assert o.build_record(Data, scid1, b"dataaa") == \
                     Data(seqnum=1, scid=b"scid", data=b"dataaa")
    assert o.build_record(Close, scid1) == \
                     Close(seqnum=2, scid=b"scid")
    assert o.build_record(Close, scid1) == \
                     Close(seqnum=3, scid=b"scid")

def test_outbound_queue():
    o, m, c = make_outbound()
    scid1 = b"scid"
    r1 = o.build_record(Open, scid1, b"proto")
    r2 = o.build_record(Data, scid1, b"data1")
    r3 = o.build_record(Data, scid1, b"data2")
    o.queue_and_send_record(r1)
    o.queue_and_send_record(r2)
    o.queue_and_send_record(r3)
    assert list(o._outbound_queue) == [r1, r2, r3]

    # we would never normally receive an ACK without first getting a
    # connection
    o.handle_ack(r2.seqnum)
    assert list(o._outbound_queue) == [r3]

    o.handle_ack(r3.seqnum)
    assert list(o._outbound_queue) == []

    o.handle_ack(r3.seqnum)  # ignored
    assert list(o._outbound_queue) == []

    o.handle_ack(r1.seqnum)  # ignored
    assert list(o._outbound_queue) == []

def test_duplicate_registerProducer():
    o, m, c = make_outbound()
    sc1 = object()
    p1 = mock.Mock()
    o.subchannel_registerProducer(sc1, p1, True)
    with pytest.raises(ValueError) as ar:
        o.subchannel_registerProducer(sc1, p1, True)
    s = str(ar.value)
    assert "registering producer" in s
    assert "before previous one" in s
    assert "was unregistered" in s

def test_connection_send_queued_unpaused():
    o, m, c = make_outbound()
    scid1 = b"scid"
    r1 = o.build_record(Open, scid1, b"proto")
    r2 = o.build_record(Data, scid1, b"data1")
    r3 = o.build_record(Data, scid1, b"data2")
    o.queue_and_send_record(r1)
    o.queue_and_send_record(r2)
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == []

    # as soon as the connection is established, everything is sent
    o.use_connection(c)
    assert c.mock_calls == [mock.call.transport.registerProducer(o, True),
                                    mock.call.send_record(r1),
                                    mock.call.send_record(r2)]
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == []
    clear_mock_calls(c)

    o.queue_and_send_record(r3)
    assert list(o._outbound_queue) == [r1, r2, r3]
    assert list(o._queued_unsent) == []
    assert c.mock_calls == [mock.call.send_record(r3)]

def test_connection_send_queued_paused():
    o, m, c = make_outbound()
    r1 = Pauser(seqnum=1)
    r2 = Pauser(seqnum=2)
    r3 = Pauser(seqnum=3)
    o.queue_and_send_record(r1)
    o.queue_and_send_record(r2)
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == []

    # pausing=True, so our mock Manager will pause the Outbound producer
    # after each write. So only r1 should have been sent before getting
    # paused
    o.use_connection(c)
    assert c.mock_calls == [mock.call.transport.registerProducer(o, True),
                                    mock.call.send_record(r1)]
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == [r2]
    clear_mock_calls(c)

    # Outbound is responsible for sending all records, so when Manager
    # wants to send a new one, and Outbound is still in the middle of
    # draining the beginning-of-connection queue, the new message gets
    # queued behind the rest (in addition to being queued in
    # _outbound_queue until an ACK retires it).
    o.queue_and_send_record(r3)
    assert list(o._outbound_queue) == [r1, r2, r3]
    assert list(o._queued_unsent) == [r2, r3]
    assert c.mock_calls == []

    o.handle_ack(r1.seqnum)
    assert list(o._outbound_queue) == [r2, r3]
    assert list(o._queued_unsent) == [r2, r3]
    assert c.mock_calls == []

def test_premptive_ack():
    # one mode I have in mind is for each side to send an immediate ACK,
    # with everything they've ever seen, as the very first message on each
    # new connection. The idea is that you might preempt sending stuff from
    # the _queued_unsent list if it arrives fast enough (in practice this
    # is more likely to be delivered via the DILATE mailbox message, but
    # the effects might be vaguely similar, so it seems worth testing
    # here). A similar situation would be if each side sends ACKs with the
    # highest seqnum they've ever seen, instead of merely ACKing the
    # message which was just received.
    o, m, c = make_outbound()
    r1 = Pauser(seqnum=1)
    r2 = Pauser(seqnum=2)
    r3 = Pauser(seqnum=3)
    o.queue_and_send_record(r1)
    o.queue_and_send_record(r2)
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == []

    o.use_connection(c)
    assert c.mock_calls == [mock.call.transport.registerProducer(o, True),
                                    mock.call.send_record(r1)]
    assert list(o._outbound_queue) == [r1, r2]
    assert list(o._queued_unsent) == [r2]
    clear_mock_calls(c)

    o.queue_and_send_record(r3)
    assert list(o._outbound_queue) == [r1, r2, r3]
    assert list(o._queued_unsent) == [r2, r3]
    assert c.mock_calls == []

    o.handle_ack(r2.seqnum)
    assert list(o._outbound_queue) == [r3]
    assert list(o._queued_unsent) == [r3]
    assert c.mock_calls == []

def test_pause():
    o, m, c = make_outbound()
    o.use_connection(c)
    assert c.mock_calls == [mock.call.transport.registerProducer(o, True)]
    assert list(o._outbound_queue) == []
    assert list(o._queued_unsent) == []
    clear_mock_calls(c)

    sc1, sc2, sc3 = object(), object(), object()
    p1, p2, p3 = mock.Mock(name="p1"), mock.Mock(
        name="p2"), mock.Mock(name="p3")

    # we aren't paused yet, since we haven't sent any data
    o.subchannel_registerProducer(sc1, p1, True)
    assert p1.mock_calls == []

    r1 = Pauser(seqnum=1)
    o.queue_and_send_record(r1)
    # now we should be paused
    assert o._paused
    assert c.mock_calls == [mock.call.send_record(r1)]
    assert p1.mock_calls == [mock.call.pauseProducing()]
    clear_mock_calls(p1, c)

    # so an IPushProducer will be paused right away
    o.subchannel_registerProducer(sc2, p2, True)
    assert p2.mock_calls == [mock.call.pauseProducing()]
    clear_mock_calls(p2)

    o.subchannel_registerProducer(sc3, p3, True)
    assert p3.mock_calls == [mock.call.pauseProducing()]
    assert o._paused_producers == {p1, p2, p3}
    assert list(o._all_producers) == [p1, p2, p3]
    clear_mock_calls(p3)

    # one resumeProducing should cause p1 to get a turn, since p2 was added
    # after we were paused and p1 was at the "end" of a one-element list.
    # If it writes anything, it will get paused again immediately.
    r2 = Pauser(seqnum=2)
    p1.resumeProducing.side_effect = lambda: c.send_record(r2)
    o.resumeProducing()
    assert p1.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p2.mock_calls == []
    assert p3.mock_calls == []
    assert c.mock_calls == [mock.call.send_record(r2)]
    clear_mock_calls(p1, p2, p3, c)
    # p2 should now be at the head of the queue
    assert list(o._all_producers) == [p2, p3, p1]

    # next turn: p2 has nothing to send, but p3 does. we should see p3
    # called but not p1. The actual sequence of expected calls is:
    # p2.resume, p3.resume, pauseProducing, set(p2.pause, p3.pause)
    r3 = Pauser(seqnum=3)
    p2.resumeProducing.side_effect = lambda: None
    p3.resumeProducing.side_effect = lambda: c.send_record(r3)
    o.resumeProducing()
    assert p1.mock_calls == []
    assert p2.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p3.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert c.mock_calls == [mock.call.send_record(r3)]
    clear_mock_calls(p1, p2, p3, c)
    # p1 should now be at the head of the queue
    assert list(o._all_producers) == [p1, p2, p3]

    # next turn: p1 has data to send, but not enough to cause a pause. same
    # for p2. p3 causes a pause
    r4 = NonPauser(seqnum=4)
    r5 = NonPauser(seqnum=5)
    r6 = Pauser(seqnum=6)
    p1.resumeProducing.side_effect = lambda: c.send_record(r4)
    p2.resumeProducing.side_effect = lambda: c.send_record(r5)
    p3.resumeProducing.side_effect = lambda: c.send_record(r6)
    o.resumeProducing()
    assert p1.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p2.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p3.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert c.mock_calls == [mock.call.send_record(r4),
                                    mock.call.send_record(r5),
                                    mock.call.send_record(r6),
                                    ]
    clear_mock_calls(p1, p2, p3, c)
    # p1 should now be at the head of the queue again
    assert list(o._all_producers) == [p1, p2, p3]

    # now we let it catch up. p1 and p2 send non-pausing data, p3 sends
    # nothing.
    r7 = NonPauser(seqnum=4)
    r8 = NonPauser(seqnum=5)
    p1.resumeProducing.side_effect = lambda: c.send_record(r7)
    p2.resumeProducing.side_effect = lambda: c.send_record(r8)
    p3.resumeProducing.side_effect = lambda: None

    o.resumeProducing()
    assert p1.mock_calls == [mock.call.resumeProducing(),
                                     ]
    assert p2.mock_calls == [mock.call.resumeProducing(),
                                     ]
    assert p3.mock_calls == [mock.call.resumeProducing(),
                                     ]
    assert c.mock_calls == [mock.call.send_record(r7),
                                    mock.call.send_record(r8),
                                    ]
    clear_mock_calls(p1, p2, p3, c)
    # p1 should now be at the head of the queue again
    assert list(o._all_producers) == [p1, p2, p3]
    assert not o._paused

    # now a producer disconnects itself (spontaneously, not from inside a
    # resumeProducing)
    o.subchannel_unregisterProducer(sc1)
    assert list(o._all_producers) == [p2, p3]
    assert p1.mock_calls == []
    assert not o._paused

    # and another disconnects itself when called
    p2.resumeProducing.side_effect = lambda: None
    p3.resumeProducing.side_effect = lambda: o.subchannel_unregisterProducer(
        sc3)
    o.pauseProducing()
    o.resumeProducing()
    assert p2.mock_calls == [mock.call.pauseProducing(),
                                     mock.call.resumeProducing()]
    assert p3.mock_calls == [mock.call.pauseProducing(),
                                     mock.call.resumeProducing()]
    clear_mock_calls(p2, p3)
    assert list(o._all_producers) == [p2]
    assert not o._paused

def test_subchannel_closed():
    o, m, c = make_outbound()

    sc1 = mock.Mock()
    p1 = mock.Mock(name="p1")
    o.subchannel_registerProducer(sc1, p1, True)
    assert p1.mock_calls == [mock.call.pauseProducing()]
    clear_mock_calls(p1)

    o.subchannel_closed(1, sc1)
    assert p1.mock_calls == []
    assert list(o._all_producers) == []

    sc2 = mock.Mock()
    o.subchannel_closed(2, sc2)

def test_disconnect():
    o, m, c = make_outbound()
    o.use_connection(c)

    sc1 = mock.Mock()
    p1 = mock.Mock(name="p1")
    o.subchannel_registerProducer(sc1, p1, True)
    assert p1.mock_calls == []
    o.stop_using_connection()
    assert p1.mock_calls == [mock.call.pauseProducing()]

def OFF_test_push_pull(self):
    # use one IPushProducer and one IPullProducer. They should take turns
    o, m, c = make_outbound()
    o.use_connection(c)
    clear_mock_calls(c)

    sc1, sc2 = object(), object()
    p1, p2 = mock.Mock(name="p1"), mock.Mock(name="p2")
    r1 = Pauser(seqnum=1)
    r2 = NonPauser(seqnum=2)

    # we aren't paused yet, since we haven't sent any data
    o.subchannel_registerProducer(sc1, p1, True)  # push
    o.queue_and_send_record(r1)
    # now we're paused
    assert o._paused
    assert c.mock_calls == [mock.call.send_record(r1)]
    assert p1.mock_calls == [mock.call.pauseProducing()]
    assert p2.mock_calls == []
    clear_mock_calls(p1, p2, c)

    p1.resumeProducing.side_effect = lambda: c.send_record(r1)
    p2.resumeProducing.side_effect = lambda: c.send_record(r2)
    o.subchannel_registerProducer(sc2, p2, False)  # pull: always ready

    # p1 is still first, since p2 was just added (at the end)
    assert o._paused
    assert c.mock_calls == []
    assert p1.mock_calls == []
    assert p2.mock_calls == []
    assert list(o._all_producers) == [p1, p2]
    clear_mock_calls(p1, p2, c)

    # resume should send r1, which should pause everything
    o.resumeProducing()
    assert o._paused
    assert c.mock_calls == [mock.call.send_record(r1),
                                    ]
    assert p1.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p2.mock_calls == []
    assert list(o._all_producers) == [p2, p1]  # now p2 is next
    clear_mock_calls(p1, p2, c)

    # next should fire p2, then p1
    o.resumeProducing()
    assert o._paused
    assert c.mock_calls == [mock.call.send_record(r2),
                                    mock.call.send_record(r1),
                                    ]
    assert p1.mock_calls == [mock.call.resumeProducing(),
                                     mock.call.pauseProducing(),
                                     ]
    assert p2.mock_calls == [mock.call.resumeProducing(),
                                     ]
    assert list(o._all_producers) == [p2, p1]  # p2 still at bat
    clear_mock_calls(p1, p2, c)

def test_pull_producer():
    # a single pull producer should write until it is paused, rate-limited
    # by the cooperator (so we'll see back-to-back resumeProducing calls
    # until the Connection is paused, or 10ms have passed, whichever comes
    # first, and if it's stopped by the timer, then the next EventualQueue
    # turn will start it off again)

    o, m, c = make_outbound()
    eq = o._test_eq
    o.use_connection(c)
    clear_mock_calls(c)
    assert not o._paused

    sc1 = mock.Mock()
    p1 = mock.Mock(name="p1")
    alsoProvides(p1, IPullProducer)

    records = [NonPauser(seqnum=1)] * 10
    records.append(Pauser(seqnum=2))
    records.append(Stopper(sc1))
    it = iter(records)
    p1.resumeProducing.side_effect = lambda: c.send_record(next(it))
    o.subchannel_registerProducer(sc1, p1, False)
    eq.flush_sync()  # fast forward into the glorious (paused) future

    assert o._paused
    assert c.mock_calls == \
                     [mock.call.send_record(r) for r in records[:-1]]
    assert p1.mock_calls == \
                     [mock.call.resumeProducing()] * (len(records) - 1)
    clear_mock_calls(c, p1)

    # next resumeProducing should cause it to disconnect
    o.resumeProducing()
    eq.flush_sync()
    assert c.mock_calls == [mock.call.send_record(records[-1])]
    assert p1.mock_calls == [mock.call.resumeProducing()]
    assert len(o._all_producers) == 0
    assert not o._paused

def test_two_pull_producers():
    # we should alternate between them until paused
    p1_records = ([NonPauser(seqnum=i) for i in range(5)] +
                  [Pauser(seqnum=5)] +
                  [NonPauser(seqnum=i) for i in range(6, 10)])
    p2_records = ([NonPauser(seqnum=i) for i in range(10, 19)] +
                  [Pauser(seqnum=19)])
    expected1 = [NonPauser(0), NonPauser(10),
                 NonPauser(1), NonPauser(11),
                 NonPauser(2), NonPauser(12),
                 NonPauser(3), NonPauser(13),
                 NonPauser(4), NonPauser(14),
                 Pauser(5)]
    expected2 = [NonPauser(15),
                 NonPauser(6), NonPauser(16),
                 NonPauser(7), NonPauser(17),
                 NonPauser(8), NonPauser(18),
                 NonPauser(9), Pauser(19),
                 ]

    o, m, c = make_outbound()
    eq = o._test_eq
    o.use_connection(c)
    clear_mock_calls(c)
    assert not o._paused

    sc1 = mock.Mock()
    p1 = mock.Mock(name="p1")
    alsoProvides(p1, IPullProducer)
    it1 = iter(p1_records)
    p1.resumeProducing.side_effect = lambda: c.send_record(next(it1))
    o.subchannel_registerProducer(sc1, p1, False)

    sc2 = mock.Mock()
    p2 = mock.Mock(name="p2")
    alsoProvides(p2, IPullProducer)
    it2 = iter(p2_records)
    p2.resumeProducing.side_effect = lambda: c.send_record(next(it2))
    o.subchannel_registerProducer(sc2, p2, False)

    eq.flush_sync()  # fast forward into the glorious (paused) future

    sends = [mock.call.resumeProducing()]
    assert o._paused
    assert c.mock_calls == \
                     [mock.call.send_record(r) for r in expected1]
    assert p1.mock_calls == 6 * sends
    assert p2.mock_calls == 5 * sends
    clear_mock_calls(c, p1, p2)

    o.resumeProducing()
    eq.flush_sync()
    assert o._paused
    assert c.mock_calls == \
                     [mock.call.send_record(r) for r in expected2]
    assert p1.mock_calls == 4 * sends
    assert p2.mock_calls == 5 * sends
    clear_mock_calls(c, p1, p2)

def test_send_if_connected():
    o, m, c = make_outbound()
    o.send_if_connected(Ack(1))  # not connected yet

    o.use_connection(c)
    o.send_if_connected(KCM())
    assert c.mock_calls == [mock.call.transport.registerProducer(o, True),
                                    mock.call.send_record(KCM())]

def test_tolerate_duplicate_pause_resume():
    o, m, c = make_outbound()
    assert o._paused  # no connection
    o.use_connection(c)
    assert not o._paused
    o.pauseProducing()
    assert o._paused
    o.pauseProducing()
    assert o._paused
    o.resumeProducing()
    assert not o._paused
    o.resumeProducing()
    assert not o._paused

def test_stopProducing():
    o, m, c = make_outbound()
    o.use_connection(c)
    assert not o._paused
    o.stopProducing()  # connection does this before loss
    assert o._paused
    o.stop_using_connection()
    assert o._paused

def test_resume_error(observe_errors):
    o, m, c = make_outbound()
    o.use_connection(c)
    sc1 = mock.Mock()
    p1 = mock.Mock(name="p1")
    alsoProvides(p1, IPullProducer)
    p1.resumeProducing.side_effect = PretendResumptionError
    o.subchannel_registerProducer(sc1, p1, False)
    o._test_eq.flush_sync()
    # the error is supposed to automatically unregister the producer
    assert list(o._all_producers) == []
    observe_errors.flush(PretendResumptionError)


def make_pushpull(pauses):
    p = mock.Mock()
    alsoProvides(p, IPullProducer)
    unregister = mock.Mock()

    clock = Clock()
    eq = EventualQueue(clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    coop = Cooperator(terminationPredicateFactory=term_factory,
                      scheduler=eq.eventually)
    pp = PullToPush(p, unregister, coop)

    it = cycle(pauses)

    def action(i):
        if isinstance(i, Exception):
            raise i
        elif i:
            pp.pauseProducing()
    p.resumeProducing.side_effect = lambda: action(next(it))
    return p, unregister, pp, eq


class PretendResumptionError(Exception):
    pass


class PretendUnregisterError(Exception):
    pass


def test_start_unpaused():
    p, unr, pp, eq = make_pushpull([True])  # pause on each resumeProducing
    # if it starts unpaused, it gets one write before being halted
    pp.startStreaming(False)
    eq.flush_sync()
    assert p.mock_calls == [mock.call.resumeProducing()] * 1
    clear_mock_calls(p)

    # now each time we call resumeProducing, we should see one delivered to
    # the underlying IPullProducer
    pp.resumeProducing()
    eq.flush_sync()
    assert p.mock_calls == [mock.call.resumeProducing()] * 1

    pp.stopStreaming()
    pp.stopStreaming()  # should tolerate this

def test_start_unpaused_two_writes():
    p, unr, pp, eq = make_pushpull([False, True])  # pause every other time
    # it should get two writes, since the first didn't pause
    pp.startStreaming(False)
    eq.flush_sync()
    assert p.mock_calls == [mock.call.resumeProducing()] * 2

def test_start_paused():
    p, unr, pp, eq = make_pushpull([True])  # pause on each resumeProducing
    pp.startStreaming(True)
    eq.flush_sync()
    assert p.mock_calls == []
    pp.stopStreaming()

def test_stop():
    p, unr, pp, eq = make_pushpull([True])
    pp.startStreaming(True)
    pp.stopProducing()
    eq.flush_sync()
    assert p.mock_calls == [mock.call.stopProducing()]


def test_error(observe_errors):
    p, unr, pp, eq = make_pushpull([PretendResumptionError()])
    unr.side_effect = lambda: pp.stopStreaming()
    pp.startStreaming(False)
    eq.flush_sync()
    assert unr.mock_calls == [mock.call()]
    observe_errors.flush(PretendResumptionError)


def test_error_during_unregister(observe_errors):
    p, unr, pp, eq = make_pushpull([PretendResumptionError()])
    unr.side_effect = PretendUnregisterError()
    pp.startStreaming(False)
    eq.flush_sync()
    assert unr.mock_calls == [mock.call()]
    observe_errors.flush(PretendResumptionError)
    observe_errors.flush(PretendUnregisterError)

    # TODO: consider making p1/p2/p3 all elements of a shared Mock, maybe I
    # could capture the inter-call ordering that way



from unittest import mock

from .._key import derive_key, derive_phase_key, encrypt_data, decrypt_data
from ..util import bytes_to_hexstr, hexstr_to_bytes
import pytest


def test_derive_errors():
    with pytest.raises(TypeError):
        derive_key(123, b"purpose")
    with pytest.raises(TypeError):
        derive_key(b"key", 123)
    with pytest.raises(TypeError):
        derive_key(b"key", b"purpose", "not len")

def test_derive_key():
    m = "588ba9eef353778b074413a0140205d90d7479e36e0dd4ee35bb729d26131ef1"
    main = hexstr_to_bytes(m)

    dk1 = derive_key(main, b"purpose1")
    assert bytes_to_hexstr(dk1) == \
                     "835b5df80ce9ca46908e8524fb308649" \
                     "122cfbcefbeaa7e65061c6ef08ee1b2a"

    dk2 = derive_key(main, b"purpose2", 10)
    assert bytes_to_hexstr(dk2) == "f2238e84315b47eb6279"

def test_derive_phase_key():
    m = "588ba9eef353778b074413a0140205d90d7479e36e0dd4ee35bb729d26131ef1"
    main = hexstr_to_bytes(m)

    dk11 = derive_phase_key(main, "side1", "phase1")
    assert bytes_to_hexstr(dk11) == \
                     "3af6a61d1a111225cc8968c6ca6265ef" \
                     "e892065c3ab46de79dda21306b062990"

    dk12 = derive_phase_key(main, "side1", "phase2")
    assert bytes_to_hexstr(dk12) == \
                     "88a1dd12182d989ff498022a9656d1e2" \
                     "806f17328d8bf5d8d0c9753e4381a752"

    dk21 = derive_phase_key(main, "side2", "phase1")
    assert bytes_to_hexstr(dk21) == \
                     "a306627b436ec23bdae3af8fa90c9ac9" \
                     "27780d86be1831003e7f617c518ea689"

    dk22 = derive_phase_key(main, "side2", "phase2")
    assert bytes_to_hexstr(dk22) == \
                     "bf99e3e16420f2dad33f9b1ccb0be146" \
                     "2b253d639dacdb50ed9496fa528d8758"


def test_encrypt():
    k = "ddc543ef8e4629a603d39dd0307a51bb1e7adb9cb259f6b085c91d0842a18679"
    key = hexstr_to_bytes(k)
    plaintext = hexstr_to_bytes("edc089a518219ec1cee184e89d2d37af")
    assert len(plaintext) == 16
    nonce = hexstr_to_bytes("2d5e43eb465aa42e750f991e425bee48"
                            "5f06abad7e04af80")
    assert len(nonce) == 24
    with mock.patch("nacl.utils.random", return_value=nonce):
        encrypted = encrypt_data(key, plaintext)
    assert len(encrypted) == 24 + 16 + 16
    assert bytes_to_hexstr(encrypted) == \
                     "2d5e43eb465aa42e750f991e425bee48" \
                     "5f06abad7e04af80fe318e39d0e4ce93" \
                     "2d2b54b300c56d2cda55ee5f0488d63e" \
                     "b1d5f76f7919a49a"

def test_decrypt():
    k = "ddc543ef8e4629a603d39dd0307a51bb1e7adb9cb259f6b085c91d0842a18679"
    key = hexstr_to_bytes(k)
    encrypted = hexstr_to_bytes("2d5e43eb465aa42e750f991e425bee48"
                                "5f06abad7e04af80fe318e39d0e4ce93"
                                "2d2b54b300c56d2cda55ee5f0488d63e"
                                "b1d5f76f7919a49a")

    decrypted = decrypt_data(key, encrypted)
    assert len(decrypted) == len(encrypted) - 24 - 16
    assert bytes_to_hexstr(decrypted) == \
                     "edc089a518219ec1cee184e89d2d37af"



import io
from collections import namedtuple
from unittest import mock
from twisted.internet import endpoints, reactor
from .._hints import (endpoint_from_hint_obj, parse_hint_argv, parse_tcp_v1_hint,
                      describe_hint_obj, parse_hint, encode_hint,
                      DirectTCPV1Hint, TorTCPV1Hint, RelayV1Hint)
import pytest

UnknownHint = namedtuple("UnknownHint", ["stuff"])


def test_endpoint_from_hint_obj():
    def efho(hint, tor=None):
        return endpoint_from_hint_obj(hint, tor, reactor)
    assert isinstance(efho(DirectTCPV1Hint("host", 1234, 0.0)), endpoints.HostnameEndpoint)
    assert efho("unknown:stuff:yowza:pivlor") is None

    # tor=None
    assert efho(TorTCPV1Hint("host", "port", 0)) is None
    assert efho(UnknownHint("foo")) is None

    tor = mock.Mock()

    def tor_ep(hostname, port):
        if hostname == "non-public":
            raise ValueError
        return ("tor_ep", hostname, port)
    tor.stream_via = mock.Mock(side_effect=tor_ep)

    assert efho(DirectTCPV1Hint("host", 1234, 0.0), tor) == \
                     ("tor_ep", "host", 1234)
    assert efho(TorTCPV1Hint("host2.onion", 1234, 0.0), tor) == \
                     ("tor_ep", "host2.onion", 1234)
    assert efho(DirectTCPV1Hint("non-public", 1234, 0.0), tor) is None

    assert efho(UnknownHint("foo"), tor) is None

def test_comparable():
    h1 = DirectTCPV1Hint("hostname", "port1", 0.0)
    h1b = DirectTCPV1Hint("hostname", "port1", 0.0)
    h2 = DirectTCPV1Hint("hostname", "port2", 0.0)
    r1 = RelayV1Hint(tuple(sorted([h1, h2])))
    r2 = RelayV1Hint(tuple(sorted([h2, h1])))
    r3 = RelayV1Hint(tuple(sorted([h1b, h2])))
    assert r1 == r2
    assert r2 == r3
    assert len({r1, r2, r3}) == 1

def test_parse_tcp_v1_hint():
    p = parse_tcp_v1_hint
    assert p({"type": "unknown"}) is None
    h = p({"type": "direct-tcp-v1", "hostname": "foo", "port": 1234})
    assert h == DirectTCPV1Hint("foo", 1234, 0.0)
    h = p({
        "type": "direct-tcp-v1",
        "hostname": "foo",
        "port": 1234,
        "priority": 2.5
    })
    assert h == DirectTCPV1Hint("foo", 1234, 2.5)
    h = p({"type": "tor-tcp-v1", "hostname": "foo", "port": 1234})
    assert h == TorTCPV1Hint("foo", 1234, 0.0)
    h = p({
        "type": "tor-tcp-v1",
        "hostname": "foo",
        "port": 1234,
        "priority": 2.5
    })
    assert h == TorTCPV1Hint("foo", 1234, 2.5)
    assert p({
        "type": "direct-tcp-v1"
    }) is None  # missing hostname
    assert p({
        "type": "direct-tcp-v1",
        "hostname": 12
    }) is None  # invalid hostname
    assert p({
            "type": "direct-tcp-v1",
            "hostname": "foo"
        }) is None  # missing port
    assert p({
            "type": "direct-tcp-v1",
            "hostname": "foo",
            "port": "not a number"
        }) is None  # invalid port

def test_parse_hint():
    p = parse_hint
    assert p({"type": "direct-tcp-v1",
                        "hostname": "foo",
                        "port": 12}) == \
                     DirectTCPV1Hint("foo", 12, 0.0)
    assert p({"type": "relay-v1",
                        "hints": [
                            {"type": "direct-tcp-v1",
                             "hostname": "foo",
                             "port": 12},
                            {"type": "unrecognized"},
                            {"type": "direct-tcp-v1",
                             "hostname": "bar",
                             "port": 13}]}) == \
                     RelayV1Hint([DirectTCPV1Hint("foo", 12, 0.0),
                                  DirectTCPV1Hint("bar", 13, 0.0)])

def test_parse_hint_argv():
    def p(hint):
        stderr = io.StringIO()
        value = parse_hint_argv(hint, stderr=stderr)
        return value, stderr.getvalue()

    h, stderr = p("tcp:host:1234")
    assert h == DirectTCPV1Hint("host", 1234, 0.0)
    assert stderr == ""

    h, stderr = p("tcp:host:1234:priority=2.6")
    assert h == DirectTCPV1Hint("host", 1234, 2.6)
    assert stderr == ""

    h, stderr = p("tcp:host:1234:unknown=stuff")
    assert h == DirectTCPV1Hint("host", 1234, 0.0)
    assert stderr == ""

    h, stderr = p("$!@#^")
    assert h is None
    assert stderr == "unparseable hint '$!@#^'\n"

    h, stderr = p("unknown:stuff")
    assert h is None
    assert stderr == \
                     "unknown hint type 'unknown' in 'unknown:stuff'\n"

    h, stderr = p("tcp:just-a-hostname")
    assert h is None
    assert stderr == \
        "unparseable TCP hint (need more colons) 'tcp:just-a-hostname'\n"

    h, stderr = p("tcp:host:number")
    assert h is None
    assert stderr == \
                     "non-numeric port in TCP hint 'tcp:host:number'\n"

    h, stderr = p("tcp:host:1234:priority=bad")
    assert h is None
    assert stderr == \
        "non-float priority= in TCP hint 'tcp:host:1234:priority=bad'\n"

    h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]")
    assert h is None
    assert stderr == "non-numeric port in TCP hint 'tcp:[2001:0db8:85a3::8a2e:0370:7334]'\n"

    h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]:1234")
    assert h == DirectTCPV1Hint(
        "2001:0db8:85a3::8a2e:0370:7334", 1234, 0.0)
    assert stderr == ""

    h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]:1234:priority=2.6")
    assert h == DirectTCPV1Hint(
        "2001:0db8:85a3::8a2e:0370:7334", 1234, 2.6)
    assert stderr == ""

    h, stderr = p("tcp:[abc::xyz]:1234")
    assert h is None
    assert stderr == "invalid IPv6 address in TCP hint 'tcp:[abc::xyz]:1234'\n"

def test_describe_hint_obj():
    d = describe_hint_obj
    assert d(DirectTCPV1Hint("host", 1234, 0.0), False, False) == \
                     "->tcp:host:1234"
    assert d(DirectTCPV1Hint("host", 1234, 0.0), True, False) == \
                     "->relay:tcp:host:1234"
    assert d(DirectTCPV1Hint("host", 1234, 0.0), False, True) == \
                     "tor->tcp:host:1234"
    assert d(DirectTCPV1Hint("host", 1234, 0.0), True, True) == \
                     "tor->relay:tcp:host:1234"
    assert d(TorTCPV1Hint("host", 1234, 0.0), False, False) == \
                     "->tor:host:1234"
    assert d(UnknownHint("stuff"), False, False) == \
                     f"->{str(UnknownHint('stuff'))}"

def test_encode_hint():
    e = encode_hint
    assert e(DirectTCPV1Hint("host", 1234, 1.0)) == \
                     {"type": "direct-tcp-v1",
                      "priority": 1.0,
                      "hostname": "host",
                      "port": 1234}
    assert e(RelayV1Hint([DirectTCPV1Hint("foo", 12, 0.0),
                                    DirectTCPV1Hint("bar", 13, 0.0)])) == \
                     {"type": "relay-v1",
                      "hints": [
                          {"type": "direct-tcp-v1",
                           "hostname": "foo",
                           "port": 12,
                           "priority": 0.0},
                          {"type": "direct-tcp-v1",
                           "hostname": "bar",
                           "port": 13,
                           "priority": 0.0},
                      ]}
    assert e(TorTCPV1Hint("host", 1234, 1.0)) == \
                     {"type": "tor-tcp-v1",
                      "priority": 1.0,
                      "hostname": "host",
                      "port": 1234}
    with pytest.raises(ValueError) as f:
        e("not a Hint")
    assert "unknown hint type" in str(f.value)
    assert "not a Hint" in str(f.value)



import re
from unittest import mock
from twisted.internet import reactor
from twisted.internet.task import Cooperator
from twisted.internet.defer import Deferred
from zope.interface import implementer

import pytest
from pytest_twisted import ensureDeferred

from ... import _interfaces
from ...eventual import EventualQueue
from ..._interfaces import ITerminator
from ..._dilation import manager
from ..._dilation._noise import NoiseConnection


@implementer(_interfaces.ISend)
class MySend:
    def __init__(self, side):
        self.rx_phase = 0
        self.side = side

    def send(self, phase, plaintext):
        # print("SEND[%s]" % self.side, phase, plaintext)
        self.peer.got(phase, plaintext)

    def got(self, phase, plaintext):
        d_mo = re.search(r'^dilate-(\d+)$', phase)
        p = int(d_mo.group(1))
        assert p == self.rx_phase
        self.rx_phase += 1
        self.dilator.received_dilate(plaintext)


@implementer(ITerminator)
class FakeTerminator:
    def __init__(self):
        self.d = Deferred()

    def stoppedD(self):
        self.d.callback(None)


@ensureDeferred
@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
async def test1():
    # print()
    send_left = MySend("left")
    send_right = MySend("right")
    send_left.peer = send_right
    send_right.peer = send_left
    key = b"\x00"*32
    eq = EventualQueue(reactor)
    cooperator = Cooperator(scheduler=eq.eventually)

    t_left = FakeTerminator()
    t_right = FakeTerminator()

    d_left = manager.Dilator(reactor, eq, cooperator, ["ged"])
    d_left.wire(send_left, t_left)
    d_left.got_key(key)
    d_left.got_wormhole_versions({"can-dilate": ["ged"]})
    send_left.dilator = d_left

    d_right = manager.Dilator(reactor, eq, cooperator, ["ged"])
    d_right.wire(send_right, t_right)
    d_right.got_key(key)
    d_right.got_wormhole_versions({"can-dilate": ["ged"]})
    send_right.dilator = d_right

    with mock.patch("wormhole._dilation.connector.ipaddrs.find_addresses",
                    return_value=["127.0.0.1"]):
        d_left.dilate(no_listen=True)
        d_right.dilate()

    # print("left connected", eps_left)
    # print("right connected", eps_right)

    # we normally shut down with w.close(), which calls Dilator.stop(),
    # which calls Terminator.stoppedD(), which (after everything else is
    # done) calls Boss.stopped
    d_left.stop()
    d_right.stop()

    await t_left.d
    await t_right.d



from unittest import mock
from zope.interface import alsoProvides
from twisted.internet.task import Clock
from twisted.python.failure import Failure
from twisted.internet.interfaces import IProtocolFactory
from twisted.internet.protocol import Protocol, Factory
import pytest
import pytest_twisted

from ...eventual import EventualQueue
from ...observer import OneShotObserver
from ..._dilation.subchannel import (SubchannelConnectorEndpoint,
                                     SubchannelListenerEndpoint,
                                     SubchannelListeningPort,
                                     SubchannelDemultiplex,
                                     SubChannel,
                                     _WormholeAddress, SubchannelAddress)
from .common import mock_manager


class CannotDilateError(Exception):
    pass


def OFFassert_makeConnection(mock_calls):
    assert len(mock_calls) == 1
    assert mock_calls[0][0] == "makeConnection"
    assert len(mock_calls[0][1]) == 1
    return mock_calls[0][1][0]


@pytest_twisted.ensureDeferred
async def test_connector_early_succeed():
    """
    Call 'connect' on a subchannel client-style endpoint before the
    Dilation channel has been established (and then fake Dilation
    establishment)
    """
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    hostaddr = _WormholeAddress()
    peeraddr = SubchannelAddress("proto")
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelConnectorEndpoint("proto", m, hostaddr, eq)
    t = SubChannel(123, m, hostaddr, peeraddr)

    class Simple(Protocol):
        pass

    f = Factory.forProtocol(Simple)
    with mock.patch("wormhole._dilation.subchannel.SubChannel",
                    return_value=t):
        d = ep.connect(f)
        eq.flush_sync()
        # the Dilation channel has NOT become available yet, so we
        # should not have actually succeeded to connect our protocol
        # until it does
        assert not d.called

    # make the Dilation connection available
    m._main_channel.fire(None)
    eq.flush_sync()

    proto = await d
    assert isinstance(proto, Simple)


@pytest_twisted.ensureDeferred
async def test_connector_early_fail():
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    hostaddr = _WormholeAddress()
    peeraddr = SubchannelAddress("proto")
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelConnectorEndpoint("proto", m, hostaddr, eq)
    t = SubChannel(123, m, hostaddr, peeraddr)

    class Simple(Protocol):
        pass

    f = Factory.forProtocol(Simple)
    with mock.patch("wormhole._dilation.subchannel.SubChannel",
                    return_value=t):
        d = ep.connect(f)
        eq.flush_sync()
        # the Dilation channel has NOT become available yet, so we
        # should not have actually succeeded to connect our protocol
        # until it does
        assert not d.called

    # there has been some kind of error establishing the Dilation
    # channel, and so it has failed.
    m._main_channel.error(Failure(CannotDilateError()))
    eq.flush_sync()

    with pytest.raises(CannotDilateError):
        await d


@pytest_twisted.ensureDeferred
async def test_connector_late_succeed():
    """
    Same as test_connector_early_succeed except the Dilation channel
    is already available before we call .connect()
    """
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    hostaddr = _WormholeAddress()
    peeraddr = SubchannelAddress("proto")
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelConnectorEndpoint("proto", m, hostaddr, eq)
    t = SubChannel(123, m, hostaddr, peeraddr)
    # make the Dilation connection available
    m._main_channel.fire(None)

    class Simple(Protocol):
        pass

    f = Factory.forProtocol(Simple)
    with mock.patch("wormhole._dilation.subchannel.SubChannel",
                    return_value=t):
        d = ep.connect(f)
        eq.flush_sync()

    proto = await d
    assert isinstance(proto, Simple)


@pytest_twisted.ensureDeferred
async def test_connector_late_fail():
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    hostaddr = _WormholeAddress()
    peeraddr = SubchannelAddress("proto")
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelConnectorEndpoint("proto", m, hostaddr, eq)
    t = SubChannel(123, m, hostaddr, peeraddr)
    # there has been some kind of error establishing the Dilation
    # channel, and so it has failed.
    m._main_channel.error(Failure(CannotDilateError()))

    class Simple(Protocol):
        pass

    f = Factory.forProtocol(Simple)
    with mock.patch("wormhole._dilation.subchannel.SubChannel",
                    return_value=t):
        d = ep.connect(f)
        eq.flush_sync()

    eq.flush_sync()

    with pytest.raises(CannotDilateError):
        await d


# refactor code could make more testable? Demultiplex thing...
@pytest_twisted.ensureDeferred
async def test_listener_early_succeed():
    # listen, main_channel_ready, got_open, got_open
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    m._host_addr = hostaddr = _WormholeAddress()
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    demux = SubchannelDemultiplex()
    m._subprotocol_Factories = demux
    ep = SubchannelListenerEndpoint("proto", m)

    class Simple(Protocol):
        pass

    class CollectBuilds(Factory):
        protocols = []

        def buildProtocol(self, addr):
            p = Factory.buildProtocol(self, addr)
            self.protocols.append(p)
            return p

    f = CollectBuilds()
    f.protocol = Simple
    demux.register("proto", f)

    d = ep.listen(f)
    eq.flush_sync()
    assert not d.called

    m._main_channel.fire(None)
    eq.flush_sync()
    lp = await d
    assert isinstance(lp, SubchannelListeningPort)

    assert lp.getHost() == hostaddr
    # TODO: IListeningPort says we must provide this, but I don't know
    # that anyone would ever call it.
    lp.startListening()

    peeraddr1 = SubchannelAddress("proto")
    t1 = SubChannel(1234, m, hostaddr, peeraddr1)
    demux._got_open(t1, peeraddr1)

    # prior asserts here all about mocks -- basically just testing
    # that the buildProtocol was called. So now we collect that
    # ourselves, but anything else interesting we might check? Can a
    # Twisted API get the protocol from somewhere?
    assert len(CollectBuilds.protocols) == 1, "expected precisely one listener protocol"
    assert all(isinstance(p, Simple) for p in CollectBuilds.protocols)

    peeraddr2 = SubchannelAddress("proto")
    t2 = SubChannel(5678, m, hostaddr, peeraddr2)
    demux._got_open(t2, peeraddr2)

    # as above, anything else interesting we should assert?
    assert len(CollectBuilds.protocols) == 2, "expected precisely one listener protocol"
    assert all(isinstance(p, Simple) for p in CollectBuilds.protocols)

    lp.stopListening()  # TODO: should this do more?


@pytest_twisted.ensureDeferred
async def test_listener_early_fail():
    # listen, main_channel_fail
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelListenerEndpoint("proto", m)

    f = mock.Mock()
    f.subprotocol = "proto"
    alsoProvides(f, IProtocolFactory)
    p1 = mock.Mock()
    p2 = mock.Mock()
    f.buildProtocol = mock.Mock(side_effect=[p1, p2])

    d = ep.listen(f)
    eq.flush_sync()
    assert not d.called

    m._main_channel.error(Failure(CannotDilateError()))
    eq.flush_sync()
    with pytest.raises(CannotDilateError):
        await d
    assert f.buildProtocol.mock_calls == []


@pytest_twisted.ensureDeferred
async def test_listener_late_succeed():
    # listen, main_channel_ready, got_open, got_open
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    m._host_addr = hostaddr = _WormholeAddress()
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    demux = SubchannelDemultiplex()
    m._subprotocol_Factories = demux
    ep = SubchannelListenerEndpoint("proto", m)
    # XXX only real difference between "early" / "late" tests is when we call this -- use a test-variant instead?
    m._main_channel.fire(None)

    class Simple(Protocol):
        pass

    class CollectBuilds(Factory):
        protocols = []

        def buildProtocol(self, addr):
            p = Factory.buildProtocol(self, addr)
            self.protocols.append(p)
            return p

    f = CollectBuilds()
    f.protocol = Simple
    demux.register("proto", f)

    d = ep.listen(f)
    eq.flush_sync()
    lp = await d
    assert isinstance(lp, SubchannelListeningPort)

    assert lp.getHost() == hostaddr
    # TODO: IListeningPort says we must provide this, but I don't know
    # that anyone would ever call it.
    lp.startListening()

    peeraddr1 = SubchannelAddress("proto")
    t1 = SubChannel(1234, m, hostaddr, peeraddr1)
    demux._got_open(t1, peeraddr1)

    # prior asserts here all about mocks -- basically just testing
    # that the buildProtocol was called. So now we collect that
    # ourselves, but anything else interesting we might check? Can a
    # Twisted API get the protocol from somewhere?
    assert len(CollectBuilds.protocols) == 1, "expected precisely one listener protocol"
    assert all(isinstance(p, Simple) for p in CollectBuilds.protocols)

    peeraddr2 = SubchannelAddress("proto")
    t2 = SubChannel(5678, m, hostaddr, peeraddr2)
    demux._got_open(t2, peeraddr2)

    # as above, anything else interesting we should assert?
    assert len(CollectBuilds.protocols) == 2, "expected precisely one listener protocol"
    assert all(isinstance(p, Simple) for p in CollectBuilds.protocols)

    lp.stopListening()  # TODO: should this do more?


@pytest_twisted.ensureDeferred
async def test_listener_late_fail():
    # main_channel_fail, listen
    m = mock_manager()
    m.allocate_subchannel_id = mock.Mock(return_value=0)
    m._host_addr = _WormholeAddress()
    eq = EventualQueue(Clock())
    m._main_channel = OneShotObserver(eq)
    ep = SubchannelListenerEndpoint("proto", m)
    m._main_channel.error(Failure(CannotDilateError()))

    f = mock.Mock()
    f.subprotocol = "proto"
    alsoProvides(f, IProtocolFactory)
    p1 = mock.Mock()
    p2 = mock.Mock()
    f.buildProtocol = mock.Mock(side_effect=[p1, p2])

    d = ep.listen(f)
    eq.flush_sync()
    with pytest.raises(CannotDilateError):
        await d
    assert f.buildProtocol.mock_calls == []



from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

NoResult = object()


class OneShotObserver:
    def __init__(self, eventual_queue):
        self._eq = eventual_queue
        self._result = NoResult
        self._observers = []  # list of Deferreds

    def when_fired(self):
        d = Deferred()
        self._observers.append(d)
        self._maybe_call_observers()
        return d

    def fire(self, result):
        assert self._result is NoResult
        self._result = result
        self._maybe_call_observers()

    def _maybe_call_observers(self):
        if self._result is NoResult:
            return
        observers, self._observers = self._observers, []
        for d in observers:
            self._eq.eventually(d.callback, self._result)

    def error(self, f):
        # errors will override an existing result
        assert isinstance(f, Failure)
        self._result = f
        self._maybe_call_observers()

    def fire_if_not_fired(self, result):
        if self._result is NoResult:
            self.fire(result)


class SequenceObserver:
    def __init__(self, eventual_queue):
        self._eq = eventual_queue
        self._error = None
        self._results = []
        self._observers = []

    def when_next_event(self):
        d = Deferred()
        if self._error:
            self._eq.eventually(d.errback, self._error)
        elif self._results:
            result = self._results.pop(0)
            self._eq.eventually(d.callback, result)
        else:
            self._observers.append(d)
        return d

    def fire(self, result):
        if isinstance(result, Failure):
            self._error = result
            for d in self._observers:
                self._eq.eventually(d.errback, self._error)
            self._observers = []
        else:
            self._results.append(result)
            if self._observers:
                d = self._observers.pop(0)
                self._eq.eventually(d.callback, self._results.pop(0))


class EmptyableSet(set):
    # manage a set which grows and shrinks over time. Fire a Deferred the first
    # time it becomes empty after you start watching for it.

    def __init__(self, *args, **kwargs):
        self._eq = kwargs.pop("_eventual_queue")  # required
        super().__init__(*args, **kwargs)
        self._observer = None

    def when_next_empty(self):
        if not self._observer:
            self._observer = OneShotObserver(self._eq)
        return self._observer.when_fired()

    def discard(self, o):
        super().discard(o)
        if self._observer and not self:
            self._observer.fire(None)
            self._observer = None



import re

from attr import attrib, attrs, evolve
from attr.validators import instance_of, optional
from automat import MethodicalMachine
from twisted.python import log
from zope.interface import implementer

from . import _interfaces
from ._allocator import Allocator
from ._code import Code, validate_code
from ._dilation.manager import Dilator
from ._input import Input
from ._key import Key
from ._lister import Lister
from ._mailbox import Mailbox
from ._nameplate import Nameplate
from ._order import Order
from ._receive import Receive
from ._rendezvous import RendezvousConnector
from ._send import Send
from ._status import WormholeStatus, AllegedSharedKey, ConfirmedKey, Closed
from ._terminator import Terminator
from ._wordlist import PGPWordList
from .errors import (LonelyError, OnlyOneCodeError, ServerError, WelcomeError,
                     WrongPasswordError, _UnknownPhaseError)
from .util import bytes_to_dict, provides


@attrs
@implementer(_interfaces.IBoss)
class Boss:
    _W = attrib()
    _side = attrib(validator=instance_of(str))
    _url = attrib(validator=instance_of(str))
    _appid = attrib(validator=instance_of(str))
    _versions = attrib(validator=instance_of(dict))
    _client_version = attrib(validator=instance_of(tuple))
    _reactor = attrib()
    _eventual_queue = attrib()
    _cooperator = attrib()
    _journal = attrib(validator=provides(_interfaces.IJournal))
    _tor = attrib(validator=optional(provides(_interfaces.ITorManager)))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    _on_status_update = attrib(default=None)  # type should be Callable[[WormholeStatus], None]
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        # XXX can we just re-order so init_other_state is first, and
        # put this status back in init_other_state
        self._current_wormhole_status = WormholeStatus()
        self._build_workers()
        self._init_other_state()
        # make sure our listener gets the "initial" state; normally we
        # only send updates when we evolve() away from this initial
        # state
        if self._on_status_update is not None:
            self._on_status_update(self._current_wormhole_status)

    def _build_workers(self):
        self._N = Nameplate(self._evolve_wormhole_status)
        self._M = Mailbox(self._side)
        self._S = Send(self._side, self._timing)
        self._O = Order(self._side, self._timing)
        self._K = Key(self._appid, self._versions, self._side, self._timing)
        self._R = Receive(self._side, self._timing)
        self._RC = RendezvousConnector(self._url, self._appid, self._side,
                                       self._reactor, self._journal, self._tor,
                                       self._timing, self._client_version, self._evolve_wormhole_status)
        self._L = Lister(self._timing)
        self._A = Allocator(self._timing)
        self._I = Input(self._timing)
        self._C = Code(self._timing)
        self._T = Terminator()
        self._D = Dilator(
            self._reactor,
            self._eventual_queue,
            self._cooperator,
            self._versions.get("can-dilate", []),
        )

        self._N.wire(self._M, self._I, self._RC, self._T)
        self._M.wire(self._N, self._RC, self._O, self._T)
        self._S.wire(self._M)
        self._O.wire(self._K, self._R)
        self._K.wire(self, self._M, self._R)
        self._R.wire(self, self._S)
        self._RC.wire(self, self._N, self._M, self._A, self._L, self._T)
        self._L.wire(self._RC, self._I)
        self._A.wire(self._RC, self._C)
        self._I.wire(self._C, self._L)
        self._C.wire(self, self._A, self._N, self._K, self._I)
        self._T.wire(self, self._RC, self._N, self._M, self._D)
        self._D.wire(self._S, self._T)

    def _init_other_state(self):
        self._did_start_code = False
        self._next_tx_phase = 0
        self._next_rx_phase = 0
        self._rx_phases = {}  # phase -> plaintext

        self._next_rx_dilate_seqnum = 0
        self._rx_dilate_seqnums = {}  # seqnum -> plaintext

        self._result = "empty"

    def _evolve_wormhole_status(self, **kwargs):
        # we have to track "the wormhole status" somewhere, because
        # we'll be connected to the Mailbox (and maybe even the peer)
        # before anyone asks for Dilation at all
        status = evolve(self._current_wormhole_status, **kwargs)
        if self._on_status_update is not None:
            self._on_status_update(status)
        # ...and so we might not even _have_ anything Dilation related yet
        if hasattr(self, "_D") and self._D._manager is not None:
            self._D._manager._wormhole_status(status)
        self._current_wormhole_status = status

    # these methods are called from outside
    def start(self):
        self._RC.start()

    def _print_trace(self, old_state, input, new_state, client_name, machine,
                     file):
        if new_state:
            print(
                f"{client_name}.{machine}[{old_state}].{input} -> [{new_state}]",
                file=file)
        else:
            # the RendezvousConnector emits message events as if
            # they were state transitions, except that old_state
            # and new_state are empty strings. "input" is one of
            # R.connected, R.rx(type phase+side), R.tx(type
            # phase), R.lost .
            print(f"{client_name}.{machine}.{input}", file=file)
        file.flush()

        def output_tracer(output):
            print(f" {client_name}.{machine}.{output}()", file=file)
            file.flush()

        return output_tracer

    def _set_trace(self, client_name, which, file):
        names = {
            "B": self,
            "N": self._N,
            "M": self._M,
            "S": self._S,
            "O": self._O,
            "K": self._K,
            "SK": self._K._SK,
            "R": self._R,
            "RC": self._RC,
            "L": self._L,
            "A": self._A,
            "I": self._I,
            "C": self._C,
            "T": self._T
        }
        for machine in which.split():
            def tracer(old_state, input, new_state, machine=machine):
                self._print_trace(
                    old_state, input, new_state,
                    client_name=client_name,
                    machine=machine,
                    file=file,
                )
            names[machine].set_trace(tracer)
            if machine == "I":
                self._I.set_debug(tracer)

    # def serialize(self):
    #     raise NotImplemented

    # and these are the state-machine transition functions, which don't take
    # args
    @m.state(initial=True)
    def S0_empty(self):
        pass  # pragma: no cover

    @m.state()
    def S1_lonely(self):
        pass  # pragma: no cover

    @m.state()
    def S2_happy(self):
        pass  # pragma: no cover

    @m.state()
    def S3_closing(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S4_closed(self):
        pass  # pragma: no cover

    # from the Wormhole

    # input/allocate/set_code are regular methods, not state-transition
    # inputs. We expect them to be called just after initialization, while
    # we're in the S0_empty state. You must call exactly one of them, and the
    # call must happen while we're in S0_empty, which makes them good
    # candidates for being a proper @m.input, but set_code() will immediately
    # (reentrantly) cause self.got_code() to be fired, which is messy. These
    # are all passthroughs to the Code machine, so one alternative would be
    # to have Wormhole call Code.{input,allocate,set_code} instead, but that
    # would require the Wormhole to be aware of Code (whereas right now
    # Wormhole only knows about this Boss instance, and everything else is
    # hidden away).
    def input_code(self):
        if self._did_start_code:
            raise OnlyOneCodeError()
        self._did_start_code = True
        return self._C.input_code()

    def allocate_code(self, code_length):
        if self._did_start_code:
            raise OnlyOneCodeError()
        self._did_start_code = True
        wl = PGPWordList()
        self._C.allocate_code(code_length, wl)

    def set_code(self, code):
        validate_code(code)  # can raise KeyFormatError
        if self._did_start_code:
            raise OnlyOneCodeError()
        self._did_start_code = True
        self._C.set_code(code)

    def dilate(self, transit_relay_location=None, no_listen=False, on_status_update=None, ping_interval=None, expected_subprotocols=None):
        # returns DilatedWormhole instance; see wormhole.dilate() docs
        return self._D.dilate(
            transit_relay_location,
            no_listen=no_listen,
            wormhole_status=self._current_wormhole_status,
            status_update=on_status_update,
            ping_interval=ping_interval,
            expected_subprotocols=expected_subprotocols,
        )

    @m.input()
    def send(self, plaintext):
        pass

    @m.input()
    def close(self):
        pass

    # from RendezvousConnector:
    # * "rx_welcome" is the Welcome message, which might signal an error, or
    #   our welcome_handler might signal one
    # * "rx_error" is error message from the server (probably because of
    #   something we said badly, or due to CrowdedError)
    # * "error" is when an exception happened while it tried to deliver
    #   something else
    def rx_welcome(self, welcome):
        try:
            if "error" in welcome:
                raise WelcomeError(welcome["error"])
            # TODO: it'd be nice to not call the handler when we're in
            # S3_closing or S4_closed states. I tried to implement this with
            # rx_welcome as an @input, but in the error case I'd be
            # delivering a new input (rx_error or something) while in the
            # middle of processing the rx_welcome input, and I wasn't sure
            # Automat would handle that correctly.
            self._W.got_welcome(welcome)  # TODO: let this raise WelcomeError?
        except WelcomeError as welcome_error:
            self.rx_unwelcome(welcome_error)

    @m.input()
    def rx_unwelcome(self, welcome_error):
        pass

    @m.input()
    def rx_error(self, errmsg, orig):
        pass

    @m.input()
    def error(self, err):
        pass

    # from Code (provoked by input/allocate/set_code)
    @m.input()
    def got_code(self, code):
        pass

    # Key sends (got_key, scared)
    # Receive sends (got_message, happy, got_verifier, scared)
    @m.input()
    def happy(self):
        pass

    @m.input()
    def scared(self):
        pass

    def got_message(self, phase, plaintext):
        assert isinstance(phase, str), type(phase)
        assert isinstance(plaintext, bytes), type(plaintext)
        d_mo = re.search(r'^dilate-(\d+)$', phase)
        if phase == "version":
            self._got_version(plaintext)
        elif d_mo:
            self._got_dilate(int(d_mo.group(1)), plaintext)
        elif re.search(r'^\d+$', phase):
            self._got_phase(int(phase), plaintext)
        else:
            # Ignore unrecognized phases, for forwards-compatibility. Use
            # log.err so tests will catch surprises.
            log.err(_UnknownPhaseError(f"received unknown phase '{phase}'"))

    @m.input()
    def _got_version(self, plaintext):
        pass

    @m.input()
    def _got_phase(self, phase, plaintext):
        pass

    @m.input()
    def _got_dilate(self, seqnum, plaintext):
        pass

    @m.input()
    def got_key(self, key):
        pass

    @m.input()
    def got_verifier(self, verifier):
        pass

    # Terminator sends closed
    @m.input()
    def closed(self):
        pass

    @m.output()
    def do_got_code(self, code):
        self._W.got_code(code)

    @m.output()
    def process_version(self, plaintext):
        # most of this is wormhole-to-wormhole, ignored for now
        # in the future, this is how Dilation is signalled
        self._their_versions = bytes_to_dict(plaintext)
        self._D.got_wormhole_versions(self._their_versions)
        # but this part is app-to-app
        app_versions = self._their_versions.get("app_versions", {})
        self._W.got_versions(app_versions)

    @m.output()
    def S_send(self, plaintext):
        assert isinstance(plaintext, bytes), type(plaintext)
        phase = self._next_tx_phase
        self._next_tx_phase += 1
        self._S.send("%d" % phase, plaintext)

    @m.output()
    def close_unwelcome(self, welcome_error):
        # assert isinstance(err, WelcomeError)
        self._result = welcome_error
        self._T.close("unwelcome")

    @m.output()
    def close_error(self, errmsg, orig):
        self._result = ServerError(errmsg)
        self._T.close("errory")

    @m.output()
    def close_scared(self):
        self._result = WrongPasswordError()
        self._T.close("scary")

    @m.output()
    def close_lonely(self):
        self._result = LonelyError()
        self._T.close("lonely")

    @m.output()
    def close_happy(self):
        self._result = "happy"
        self._T.close("happy")

    @m.output()
    def W_got_key(self, key):
        self._W.got_key(key)

    @m.output()
    def D_got_key(self, key):
        self._D.got_key(key)

    @m.output()
    def send_status_peer_key(self, key):
        self._evolve_wormhole_status(
            peer_key=AllegedSharedKey(),
        )

    @m.output()
    def send_status_confirmed_key(self, plaintext):
        self._evolve_wormhole_status(
            peer_key=ConfirmedKey(),
        )

    @m.output()
    def send_status_closed(self):
        self._evolve_wormhole_status(
            mailbox_connection=Closed(),
        )

    @m.output()
    def W_got_verifier(self, verifier):
        self._W.got_verifier(verifier)

    @m.output()
    def W_received(self, phase, plaintext):
        assert isinstance(phase, int), type(phase)
        # we call Wormhole.received() in strict phase order, with no gaps
        self._rx_phases[phase] = plaintext
        while self._next_rx_phase in self._rx_phases:
            self._W.received(self._rx_phases.pop(self._next_rx_phase))
            self._next_rx_phase += 1

    @m.output()
    def D_received_dilate(self, seqnum, plaintext):
        assert isinstance(seqnum, int), type(seqnum)
        # strict phase order, no gaps
        self._rx_dilate_seqnums[seqnum] = plaintext
        while self._next_rx_dilate_seqnum in self._rx_dilate_seqnums:
            m = self._rx_dilate_seqnums.pop(self._next_rx_dilate_seqnum)
            self._D.received_dilate(m)
            self._next_rx_dilate_seqnum += 1

    @m.output()
    def W_close_with_error(self, err):
        self._result = err  # exception
        self._W.closed(self._result)

    @m.output()
    def W_closed(self):
        # result is either "happy" or a WormholeError of some sort
        self._W.closed(self._result)

    S0_empty.upon(close, enter=S3_closing, outputs=[close_lonely])
    S0_empty.upon(send, enter=S0_empty, outputs=[S_send])
    S0_empty.upon(rx_unwelcome, enter=S3_closing, outputs=[close_unwelcome])
    S0_empty.upon(got_code, enter=S1_lonely, outputs=[do_got_code])
    S0_empty.upon(rx_error, enter=S3_closing, outputs=[close_error])
    S0_empty.upon(error, enter=S4_closed, outputs=[W_close_with_error, send_status_closed])

    S1_lonely.upon(rx_unwelcome, enter=S3_closing, outputs=[close_unwelcome])
    S1_lonely.upon(happy, enter=S2_happy, outputs=[])
    S1_lonely.upon(scared, enter=S3_closing, outputs=[close_scared])
    S1_lonely.upon(close, enter=S3_closing, outputs=[close_lonely])
    S1_lonely.upon(send, enter=S1_lonely, outputs=[S_send])
    S1_lonely.upon(got_key, enter=S1_lonely, outputs=[W_got_key, D_got_key, send_status_peer_key])
    S1_lonely.upon(rx_error, enter=S3_closing, outputs=[close_error])
    S1_lonely.upon(error, enter=S4_closed, outputs=[W_close_with_error, send_status_closed])

    S2_happy.upon(rx_unwelcome, enter=S3_closing, outputs=[close_unwelcome])
    S2_happy.upon(got_verifier, enter=S2_happy, outputs=[W_got_verifier])
    S2_happy.upon(_got_phase, enter=S2_happy, outputs=[W_received])
    S2_happy.upon(_got_version, enter=S2_happy, outputs=[process_version, send_status_confirmed_key])
    S2_happy.upon(_got_dilate, enter=S2_happy, outputs=[D_received_dilate])
    S2_happy.upon(scared, enter=S3_closing, outputs=[close_scared])
    S2_happy.upon(close, enter=S3_closing, outputs=[close_happy])
    S2_happy.upon(send, enter=S2_happy, outputs=[S_send])
    S2_happy.upon(rx_error, enter=S3_closing, outputs=[close_error])
    S2_happy.upon(error, enter=S4_closed, outputs=[W_close_with_error, send_status_closed])

    S3_closing.upon(rx_unwelcome, enter=S3_closing, outputs=[])
    S3_closing.upon(rx_error, enter=S3_closing, outputs=[])
    S3_closing.upon(got_verifier, enter=S3_closing, outputs=[])
    S3_closing.upon(_got_phase, enter=S3_closing, outputs=[])
    S3_closing.upon(_got_version, enter=S3_closing, outputs=[])
    S3_closing.upon(_got_dilate, enter=S3_closing, outputs=[])
    S3_closing.upon(happy, enter=S3_closing, outputs=[])
    S3_closing.upon(scared, enter=S3_closing, outputs=[])
    S3_closing.upon(close, enter=S3_closing, outputs=[])
    S3_closing.upon(send, enter=S3_closing, outputs=[])
    S3_closing.upon(closed, enter=S4_closed, outputs=[W_closed, send_status_closed])
    S3_closing.upon(error, enter=S4_closed, outputs=[W_close_with_error, send_status_closed])

    S4_closed.upon(rx_unwelcome, enter=S4_closed, outputs=[])
    S4_closed.upon(got_verifier, enter=S4_closed, outputs=[])
    S4_closed.upon(_got_phase, enter=S4_closed, outputs=[])
    S4_closed.upon(_got_version, enter=S4_closed, outputs=[])
    S4_closed.upon(_got_dilate, enter=S4_closed, outputs=[])
    S4_closed.upon(happy, enter=S4_closed, outputs=[])
    S4_closed.upon(scared, enter=S4_closed, outputs=[])
    S4_closed.upon(close, enter=S4_closed, outputs=[])
    S4_closed.upon(send, enter=S4_closed, outputs=[])
    S4_closed.upon(error, enter=S4_closed, outputs=[])



from attr import attrs, attrib
from zope.interface import implementer
from twisted.python import log
from .._interfaces import IDilationManager, IInbound, ISubChannel
from ..util import provides
from .subchannel import (SubChannel, SubchannelAddress, UnexpectedSubprotocol)


class DuplicateOpenError(Exception):
    pass


class DataForMissingSubchannelError(Exception):
    pass


class CloseForMissingSubchannelError(Exception):
    pass


@attrs
@implementer(IInbound)
class Inbound:
    # Inbound flow control: TCP delivers data to Connection.dataReceived,
    # Connection delivers to our handle_data, we deliver to
    # SubChannel.remote_data, subchannel delivers to proto.dataReceived
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib()

    def __attrs_post_init__(self):
        # we route inbound Data records to Subchannels .dataReceived
        self._open_subchannels = {}  # scid -> Subchannel
        self._paused_subchannels = set()  # Subchannels that have paused us
        # the set is non-empty, we pause the transport
        self._highest_inbound_acked = -1
        self._connection = None

    # from our Manager
#    def set_listener_endpoint(self, listener_endpoint):
#        self._listener_endpoint = listener_endpoint

    def use_connection(self, c):
        self._connection = c
        # We can pause the connection's reads when we receive too much data. If
        # this is a non-initial connection, then we might already have
        # subchannels that are paused from before, so we might need to pause
        # the new connection before it can send us any data
        if self._paused_subchannels:
            self._connection.pauseProducing()

    def subchannel_local_open(self, scid, sc):
        assert ISubChannel.providedBy(sc)
        assert scid not in self._open_subchannels
        self._open_subchannels[scid] = sc

    # Inbound is responsible for tracking the high watermark and deciding
    # whether to ignore inbound messages or not

    def is_record_old(self, r):
        if r.seqnum <= self._highest_inbound_acked:
            return True
        return False

    def update_ack_watermark(self, seqnum):
        self._highest_inbound_acked = max(self._highest_inbound_acked,
                                          seqnum)

    def handle_open(self, scid, subprotocol):
        log.msg("inbound.handle_open", scid, subprotocol)
        if scid in self._open_subchannels:
            log.err(DuplicateOpenError(
                f"received duplicate OPEN for {scid}"))
            return
        peer_addr = SubchannelAddress(subprotocol)
        sc = SubChannel(scid, self._manager, self._host_addr, peer_addr)
        self._open_subchannels[scid] = sc
        # this can produce a (synchronous) UnexpectedSubprotocol if
        # the user specified "expected subprotocols" but this one
        # isn't in the list.
        try:
            # maybe this should be in Manager?
            self._manager._subprotocol_factories._got_open(sc, peer_addr)
        except UnexpectedSubprotocol:
            self._manager.send_close(scid)
            del self._open_subchannels[scid]

    def handle_data(self, scid, data):
        log.msg("inbound.handle_data", scid, len(data))
        sc = self._open_subchannels.get(scid)
        if sc is None:
            log.err(DataForMissingSubchannelError(
                f"received DATA for non-existent subchannel {scid}"))
            return
        sc.remote_data(data)

    def handle_close(self, scid):
        log.msg("inbound.handle_close", scid)
        sc = self._open_subchannels.get(scid)
        if sc is None:
            log.err(CloseForMissingSubchannelError(
                f"received CLOSE for non-existent subchannel {scid}"))
            return
        sc.remote_close()

    def subchannel_closed(self, scid, sc):
        # connectionLost has just been signalled
        assert self._open_subchannels[scid] is sc
        del self._open_subchannels[scid]

    def stop_using_connection(self):
        self._connection = None

    # from our Subchannel, or rather from the Protocol above it and sent
    # through the subchannel

    # The subchannel is an IProducer, and application protocols can always
    # thell them to pauseProducing if we're delivering inbound data too
    # quickly. They don't need to register anything.

    def subchannel_pauseProducing(self, sc):
        was_paused = bool(self._paused_subchannels)
        self._paused_subchannels.add(sc)
        if self._connection and not was_paused:
            self._connection.pauseProducing()

    def subchannel_resumeProducing(self, sc):
        was_paused = bool(self._paused_subchannels)
        self._paused_subchannels.discard(sc)
        if self._connection and was_paused and not self._paused_subchannels:
            self._connection.resumeProducing()

    def subchannel_stopProducing(self, sc):
        # This protocol doesn't want any additional data. If we were a normal
        # (single-owner) Transport, we'd call .loseConnection now. But our
        # Connection is shared among many subchannels, so instead we just
        # stop letting them pause the connection.
        was_paused = bool(self._paused_subchannels)
        self._paused_subchannels.discard(sc)
        if self._connection and was_paused and not self._paused_subchannels:
            self._connection.resumeProducing()

    # TODO: we might refactor these pause/resume/stop methods by building a
    # context manager that look at the paused/not-paused state first, then
    # lets the caller modify self._paused_subchannels, then looks at it a
    # second time, and calls c.pauseProducing/c.resumeProducing as
    # appropriate. I'm not sure it would be any cleaner, though.
